from .rfc5849 import SIGNATURE_HMAC_SHA1
from .rfc5849 import SIGNATURE_PLAINTEXT
from .rfc5849 import SIGNATURE_RSA_SHA1
from .rfc5849 import SIGNATURE_TYPE_BODY
from .rfc5849 import SIGNATURE_TYPE_HEADER
from .rfc5849 import SIGNATURE_TYPE_QUERY
from .rfc5849 import AuthorizationServer
from .rfc5849 import ClientAuth
from .rfc5849 import ClientMixin
from .rfc5849 import OAuth1Request
from .rfc5849 import ResourceProtector
from .rfc5849 import TemporaryCredential
from .rfc5849 import TemporaryCredentialMixin
from .rfc5849 import TokenCredentialMixin

__all__ = [
    "OAuth1Request",
    "ClientAuth",
    "SIGNATURE_HMAC_SHA1",
    "SIGNATURE_RSA_SHA1",
    "SIGNATURE_PLAINTEXT",
    "SIGNATURE_TYPE_HEADER",
    "SIGNATURE_TYPE_QUERY",
    "SIGNATURE_TYPE_BODY",
    "ClientMixin",
    "TemporaryCredentialMixin",
    "TokenCredentialMixin",
    "TemporaryCredential",
    "AuthorizationServer",
    "ResourceProtector",
]



import time

from django.contrib.auth.models import User
from django.db.models import CASCADE
from django.db.models import CharField
from django.db.models import ForeignKey
from django.db.models import IntegerField
from django.db.models import Model
from django.db.models import TextField

from authlib.common.security import generate_token
from authlib.oauth2.rfc6749 import AuthorizationCodeMixin
from authlib.oauth2.rfc6749 import ClientMixin
from authlib.oauth2.rfc6749 import TokenMixin
from authlib.oauth2.rfc6749.util import list_to_scope
from authlib.oauth2.rfc6749.util import scope_to_list


def now_timestamp():
    return int(time.time())


class Client(Model, ClientMixin):
    user = ForeignKey(User, on_delete=CASCADE)
    client_id = CharField(max_length=48, unique=True, db_index=True)
    client_secret = CharField(max_length=48, blank=True)
    redirect_uris = TextField(default="")
    default_redirect_uri = TextField(blank=False, default="")
    scope = TextField(default="")
    response_type = TextField(default="")
    grant_type = TextField(default="")
    token_endpoint_auth_method = CharField(max_length=120, default="")

    def get_client_id(self):
        return self.client_id

    def get_default_redirect_uri(self):
        return self.default_redirect_uri

    def get_allowed_scope(self, scope):
        if not scope:
            return ""
        allowed = set(scope_to_list(self.scope))
        return list_to_scope([s for s in scope.split() if s in allowed])

    def check_redirect_uri(self, redirect_uri):
        if redirect_uri == self.default_redirect_uri:
            return True
        return redirect_uri in self.redirect_uris

    def check_client_secret(self, client_secret):
        return self.client_secret == client_secret

    def check_endpoint_auth_method(self, method, endpoint):
        if endpoint == "token":
            return self.token_endpoint_auth_method == method
        return True

    def check_response_type(self, response_type):
        allowed = self.response_type.split()
        return response_type in allowed

    def check_grant_type(self, grant_type):
        allowed = self.grant_type.split()
        return grant_type in allowed


class OAuth2Token(Model, TokenMixin):
    user = ForeignKey(User, on_delete=CASCADE)
    client_id = CharField(max_length=48, db_index=True)
    token_type = CharField(max_length=40)
    access_token = CharField(max_length=255, unique=True, null=False)
    refresh_token = CharField(max_length=255, db_index=True)
    scope = TextField(default="")

    issued_at = IntegerField(null=False, default=now_timestamp)
    expires_in = IntegerField(null=False, default=0)
    access_token_revoked_at = IntegerField(default=0)
    refresh_token_revoked_at = IntegerField(default=0)

    def check_client(self, client):
        return self.client_id == client.client_id

    def get_scope(self):
        return self.scope

    def get_expires_in(self):
        return self.expires_in

    def is_revoked(self):
        return self.access_token_revoked_at or self.refresh_token_revoked_at

    def is_expired(self):
        if not self.expires_in:
            return False

        expires_at = self.issued_at + self.expires_in
        return expires_at < time.time()

    def is_refresh_token_active(self):
        return not self.refresh_token_revoked_at


class OAuth2Code(Model, AuthorizationCodeMixin):
    user = ForeignKey(User, on_delete=CASCADE)
    client_id = CharField(max_length=48, db_index=True)
    code = CharField(max_length=120, unique=True, null=False)
    redirect_uri = TextField(default="", null=True)
    response_type = TextField(default="")
    scope = TextField(default="", null=True)
    auth_time = IntegerField(null=False, default=now_timestamp)

    def is_expired(self):
        return self.auth_time + 300 < time.time()

    def get_redirect_uri(self):
        return self.redirect_uri

    def get_scope(self):
        return self.scope or ""

    def get_auth_time(self):
        return self.auth_time


class CodeGrantMixin:
    def query_authorization_code(self, code, client):
        try:
            item = OAuth2Code.objects.get(code=code, client_id=client.client_id)
        except OAuth2Code.DoesNotExist:
            return None

        if not item.is_expired():
            return item

    def delete_authorization_code(self, authorization_code):
        authorization_code.delete()

    def authenticate_user(self, authorization_code):
        return authorization_code.user


def generate_authorization_code(client, grant_user, request, **extra):
    code = generate_token(48)
    item = OAuth2Code(
        code=code,
        client_id=client.client_id,
        redirect_uri=request.payload.redirect_uri,
        response_type=request.payload.response_type,
        scope=request.payload.scope,
        user=grant_user,
        **extra,
    )
    item.save()
    return code



SECRET_KEY = "django-secret"

DATABASES = {
    "default": {
        "ENGINE": "django.db.backends.sqlite3",
        "NAME": ":memory:",
    }
}

MIDDLEWARE = ["django.contrib.sessions.middleware.SessionMiddleware"]

SESSION_ENGINE = "django.contrib.sessions.backends.cache"

CACHES = {
    "default": {
        "BACKEND": "django.core.cache.backends.locmem.LocMemCache",
        "LOCATION": "unique-snowflake",
    }
}

INSTALLED_APPS = [
    "django.contrib.contenttypes",
    "django.contrib.auth",
    "tests.django.test_oauth1",
    "tests.django.test_oauth2",
]

AUTHLIB_OAUTH_CLIENTS = {
    "dev_overwrite": {
        "client_id": "dev-client-id",
        "client_secret": "dev-client-secret",
        "access_token_params": {"foo": "foo-1", "bar": "bar-2"},
    }
}

USE_TZ = True

# Default OAuth1 configuration for tests
AUTHLIB_OAUTH1_PROVIDER = {"signature_methods": ["PLAINTEXT", "HMAC-SHA1"]}



import time

from django.core.cache import cache
from django.test import override_settings

from authlib.oauth1.rfc5849 import signature
from tests.util import decode_response
from tests.util import read_file_path


def prepare_temporary_credential(server):
    token = {
        "oauth_token": "abc",
        "oauth_token_secret": "abc-secret",
        "oauth_verifier": "abc-verifier",
        "client_id": "client",
        "user_id": 1,
    }
    key_prefix = server._temporary_credential_key_prefix
    key = key_prefix + token["oauth_token"]
    cache.set(key, token, timeout=server._temporary_expires_in)


def test_invalid_token_request_parameters(factory, server):
    url = "/oauth/token"

    # case 1
    request = factory.post(url)
    resp = server.create_token_response(request)
    data = decode_response(resp.content)
    assert data["error"] == "missing_required_parameter"
    assert "oauth_consumer_key" in data["error_description"]

    # case 2
    request = factory.post(url, data={"oauth_consumer_key": "a"})
    resp = server.create_token_response(request)
    data = decode_response(resp.content)
    assert data["error"] == "invalid_client"

    # case 3
    request = factory.post(url, data={"oauth_consumer_key": "client"})
    resp = server.create_token_response(request)
    data = decode_response(resp.content)
    assert data["error"] == "missing_required_parameter"
    assert "oauth_token" in data["error_description"]

    # case 4
    request = factory.post(
        url, data={"oauth_consumer_key": "client", "oauth_token": "a"}
    )
    resp = server.create_token_response(request)
    data = decode_response(resp.content)
    assert data["error"] == "invalid_token"


def test_duplicated_oauth_parameters(factory, server):
    url = "/oauth/token?oauth_consumer_key=client"
    request = factory.post(
        url,
        data={
            "oauth_consumer_key": "client",
            "oauth_token": "abc",
            "oauth_verifier": "abc",
        },
    )
    resp = server.create_token_response(request)
    data = decode_response(resp.content)
    assert data["error"] == "duplicated_oauth_protocol_parameter"


@override_settings(AUTHLIB_OAUTH1_PROVIDER={"signature_methods": ["PLAINTEXT"]})
def test_plaintext_signature(factory, server):
    url = "/oauth/token"

    # case 1: success
    prepare_temporary_credential(server)
    auth_header = (
        'OAuth oauth_consumer_key="client",'
        'oauth_signature_method="PLAINTEXT",'
        'oauth_token="abc",'
        'oauth_verifier="abc-verifier",'
        'oauth_signature="secret&abc-secret"'
    )
    request = factory.post(url, HTTP_AUTHORIZATION=auth_header)
    resp = server.create_token_response(request)
    data = decode_response(resp.content)
    assert "oauth_token" in data

    # case 2: invalid signature
    prepare_temporary_credential(server)
    request = factory.post(
        url,
        data={
            "oauth_consumer_key": "client",
            "oauth_signature_method": "PLAINTEXT",
            "oauth_token": "abc",
            "oauth_verifier": "abc-verifier",
            "oauth_signature": "invalid-signature",
        },
    )
    resp = server.create_token_response(request)
    data = decode_response(resp.content)
    assert data["error"] == "invalid_signature"


def test_hmac_sha1_signature(factory, server):
    url = "/oauth/token"

    params = [
        ("oauth_consumer_key", "client"),
        ("oauth_token", "abc"),
        ("oauth_verifier", "abc-verifier"),
        ("oauth_signature_method", "HMAC-SHA1"),
        ("oauth_timestamp", str(int(time.time()))),
        ("oauth_nonce", "hmac-sha1-nonce"),
    ]
    base_string = signature.construct_base_string(
        "POST", "http://testserver/oauth/token", params
    )
    sig = signature.hmac_sha1_signature(base_string, "secret", "abc-secret")
    params.append(("oauth_signature", sig))
    auth_param = ",".join([f'{k}="{v}"' for k, v in params])
    auth_header = "OAuth " + auth_param

    # case 1: success
    prepare_temporary_credential(server)
    request = factory.post(url, HTTP_AUTHORIZATION=auth_header)
    resp = server.create_token_response(request)
    data = decode_response(resp.content)
    assert "oauth_token" in data

    # case 2: exists nonce
    prepare_temporary_credential(server)
    request = factory.post(url, HTTP_AUTHORIZATION=auth_header)
    resp = server.create_token_response(request)
    data = decode_response(resp.content)
    assert data["error"] == "invalid_nonce"


def test_rsa_sha1_signature(factory, rsa_server):
    url = "/oauth/token"
    server = rsa_server

    prepare_temporary_credential(server)
    params = [
        ("oauth_consumer_key", "client"),
        ("oauth_token", "abc"),
        ("oauth_verifier", "abc-verifier"),
        ("oauth_signature_method", "RSA-SHA1"),
        ("oauth_timestamp", str(int(time.time()))),
        ("oauth_nonce", "rsa-sha1-nonce"),
    ]
    base_string = signature.construct_base_string(
        "POST", "http://testserver/oauth/token", params
    )
    sig = signature.rsa_sha1_signature(base_string, read_file_path("rsa_private.pem"))
    params.append(("oauth_signature", sig))
    auth_param = ",".join([f'{k}="{v}"' for k, v in params])
    auth_header = "OAuth " + auth_param

    request = factory.post(url, HTTP_AUTHORIZATION=auth_header)
    resp = server.create_token_response(request)
    data = decode_response(resp.content)
    assert "oauth_token" in data

    # case: invalid signature
    prepare_temporary_credential(server)
    auth_param = auth_param.replace("rsa-sha1-nonce", "alt-sha1-nonce")
    auth_header = "OAuth " + auth_param
    request = factory.post(url, HTTP_AUTHORIZATION=auth_header)
    resp = server.create_token_response(request)
    data = decode_response(resp.content)
    assert data["error"] == "invalid_signature"



import pytest

from authlib.jose import JsonWebEncryption
from authlib.jose import OctKey
from authlib.jose.drafts import register_jwe_draft

register_jwe_draft(JsonWebEncryption)


def test_dir_alg_c20p():
    jwe = JsonWebEncryption()
    key = OctKey.generate_key(256, is_private=True)
    protected = {"alg": "dir", "enc": "C20P"}
    data = jwe.serialize_compact(protected, b"hello", key)
    rv = jwe.deserialize_compact(data, key)
    assert rv["payload"] == b"hello"

    key2 = OctKey.generate_key(128, is_private=True)
    with pytest.raises(ValueError):
        jwe.deserialize_compact(data, key2)

    with pytest.raises(ValueError):
        jwe.serialize_compact(protected, b"hello", key2)


def test_dir_alg_xc20p():
    pytest.importorskip("Cryptodome.Cipher.ChaCha20_Poly1305")

    jwe = JsonWebEncryption()
    key = OctKey.generate_key(256, is_private=True)
    protected = {"alg": "dir", "enc": "XC20P"}
    data = jwe.serialize_compact(protected, b"hello", key)
    rv = jwe.deserialize_compact(data, key)
    assert rv["payload"] == b"hello"

    key2 = OctKey.generate_key(128, is_private=True)
    with pytest.raises(ValueError):
        jwe.deserialize_compact(data, key2)

    with pytest.raises(ValueError):
        jwe.serialize_compact(protected, b"hello", key2)


def test_xc20p_content_encryption_decryption():
    pytest.importorskip("Cryptodome.Cipher.ChaCha20_Poly1305")

    # https://datatracker.ietf.org/doc/html/draft-irtf-cfrg-xchacha-03#appendix-A.3.1
    enc = JsonWebEncryption.ENC_REGISTRY["XC20P"]

    plaintext = bytes.fromhex(
        "4c616469657320616e642047656e746c656d656e206f662074686520636c6173"
        + "73206f66202739393a204966204920636f756c64206f6666657220796f75206f"
        + "6e6c79206f6e652074697020666f7220746865206675747572652c2073756e73"
        + "637265656e20776f756c642062652069742e"
    )
    aad = bytes.fromhex("50515253c0c1c2c3c4c5c6c7")
    key = bytes.fromhex(
        "808182838485868788898a8b8c8d8e8f909192939495969798999a9b9c9d9e9f"
    )
    iv = bytes.fromhex("404142434445464748494a4b4c4d4e4f5051525354555657")

    ciphertext, tag = enc.encrypt(plaintext, aad, iv, key)
    assert ciphertext == bytes.fromhex(
        "bd6d179d3e83d43b9576579493c0e939572a1700252bfaccbed2902c21396cbb"
        + "731c7f1b0b4aa6440bf3a82f4eda7e39ae64c6708c54c216cb96b72e1213b452"
        + "2f8c9ba40db5d945b11b69b982c1bb9e3f3fac2bc369488f76b2383565d3fff9"
        + "21f9664c97637da9768812f615c68b13b52e"
    )
    assert tag == bytes.fromhex("c0875924c1c7987947deafd8780acf49")

    decrypted_plaintext = enc.decrypt(ciphertext, aad, iv, tag, key)
    assert decrypted_plaintext == plaintext



import json
import os

from authlib.common.encoding import to_unicode
from authlib.common.urls import url_decode

ROOT = os.path.abspath(os.path.dirname(__file__))


def get_file_path(name):
    return os.path.join(ROOT, "files", name)


def read_file_path(name):
    with open(get_file_path(name)) as f:
        if name.endswith(".json"):
            return json.load(f)
        return f.read()


def decode_response(data):
    return dict(url_decode(to_unicode(data)))



import os

import pytest
from flask import Flask

from tests.flask.test_oauth2.oauth2_server import create_authorization_server

from .models import Client
from .models import Token
from .models import User


@pytest.fixture(autouse=True)
def env():
    os.environ["AUTHLIB_INSECURE_TRANSPORT"] = "true"
    yield
    del os.environ["AUTHLIB_INSECURE_TRANSPORT"]


@pytest.fixture
def app():
    app = Flask(__name__)
    app.debug = True
    app.testing = True
    app.secret_key = "testing"
    app.config.update(
        {
            "SQLALCHEMY_TRACK_MODIFICATIONS": False,
            "SQLALCHEMY_DATABASE_URI": "sqlite://",
            "OAUTH2_ERROR_URIS": [
                ("invalid_client", "https://client.test/error#invalid_client")
            ],
        }
    )
    with app.app_context():
        yield app


@pytest.fixture
def db(app):
    from .models import db

    db.init_app(app)
    db.create_all()
    yield db
    db.drop_all()


@pytest.fixture
def test_client(app):
    return app.test_client()


@pytest.fixture(autouse=True)
def user(db):
    user = User(username="foo")
    db.session.add(user)
    db.session.commit()
    yield user
    db.session.delete(user)


@pytest.fixture
def client(db, user):
    client = Client(
        user_id=user.id,
        client_id="client-id",
        client_secret="client-secret",
    )
    client.set_client_metadata(
        {
            "redirect_uris": ["https://client.test/authorized"],
            "scope": "profile",
            "grant_types": ["authorization_code"],
            "response_types": ["code"],
        }
    )
    db.session.add(client)
    db.session.commit()
    yield client
    db.session.delete(client)


@pytest.fixture
def server(app):
    return create_authorization_server(app)


@pytest.fixture
def token(db):
    token = Token(
        user_id=1,
        client_id="client-id",
        token_type="bearer",
        access_token="a1",
        refresh_token="r1",
        scope="profile",
        expires_in=3600,
    )
    db.session.add(token)
    db.session.commit()
    yield token
    db.session.delete(token)



from ..base_client import BaseOAuth
from ..base_client import OAuthError
from .apps import DjangoOAuth1App
from .apps import DjangoOAuth2App
from .integration import DjangoIntegration
from .integration import token_update


class OAuth(BaseOAuth):
    oauth1_client_cls = DjangoOAuth1App
    oauth2_client_cls = DjangoOAuth2App
    framework_integration_cls = DjangoIntegration


__all__ = [
    "OAuth",
    "DjangoOAuth1App",
    "DjangoOAuth2App",
    "DjangoIntegration",
    "token_update",
    "OAuthError",
]



import time

import pytest
from flask import json

from authlib.oauth2.rfc8628 import (
    DeviceAuthorizationEndpoint as _DeviceAuthorizationEndpoint,
)
from authlib.oauth2.rfc8628 import DeviceCodeGrant as _DeviceCodeGrant
from authlib.oauth2.rfc8628 import DeviceCredentialDict

from .models import Client
from .models import User
from .models import db

device_credentials = {
    "valid-device": {
        "client_id": "client-id",
        "expires_in": 1800,
        "user_code": "code",
    },
    "expired-token": {
        "client_id": "client-id",
        "expires_in": -100,
        "user_code": "none",
    },
    "invalid-client": {
        "client_id": "invalid",
        "expires_in": 1800,
        "user_code": "none",
    },
    "denied-code": {
        "client_id": "client-id",
        "expires_in": 1800,
        "user_code": "denied",
    },
    "grant-code": {
        "client_id": "client-id",
        "expires_in": 1800,
        "user_code": "code",
    },
    "pending-code": {
        "client_id": "client-id",
        "expires_in": 1800,
        "user_code": "none",
    },
}


class DeviceCodeGrant(_DeviceCodeGrant):
    def query_device_credential(self, device_code):
        data = device_credentials.get(device_code)
        if not data:
            return None

        now = int(time.time())
        data["expires_at"] = now + data["expires_in"]
        data["device_code"] = device_code
        data["scope"] = "profile"
        data["interval"] = 5
        data["verification_uri"] = "https://resource.test/activate"
        return DeviceCredentialDict(data)

    def query_user_grant(self, user_code):
        if user_code == "code":
            return db.session.get(User, 1), True
        if user_code == "denied":
            return db.session.get(User, 1), False
        return None

    def should_slow_down(self, credential):
        return False


class DeviceAuthorizationEndpoint(_DeviceAuthorizationEndpoint):
    def get_verification_uri(self):
        return "https://resource.test/activate"

    def save_device_credential(self, client_id, scope, data):
        pass


@pytest.fixture(autouse=True)
def server(server, app):
    server.register_grant(DeviceCodeGrant)

    @app.route("/device_authorize", methods=["POST"])
    def device_authorize():
        name = DeviceAuthorizationEndpoint.ENDPOINT_NAME
        return server.create_endpoint_response(name)

    server.register_endpoint(DeviceAuthorizationEndpoint)

    return server


@pytest.fixture(autouse=True)
def client(client, db):
    client.set_client_metadata(
        {
            "redirect_uris": ["https://client.test/authorized"],
            "scope": "profile",
            "grant_types": [DeviceCodeGrant.GRANT_TYPE],
            "token_endpoint_auth_method": "none",
        }
    )
    db.session.add(client)
    db.session.commit()
    return client


def test_invalid_request(test_client):
    rv = test_client.post(
        "/oauth/token",
        data={
            "grant_type": DeviceCodeGrant.GRANT_TYPE,
            "client_id": "test",
        },
    )
    resp = json.loads(rv.data)
    assert resp["error"] == "invalid_request"

    rv = test_client.post(
        "/oauth/token",
        data={
            "grant_type": DeviceCodeGrant.GRANT_TYPE,
            "device_code": "missing",
            "client_id": "client-id",
        },
    )
    resp = json.loads(rv.data)
    assert resp["error"] == "invalid_request"


def test_unauthorized_client(test_client, db, client):
    rv = test_client.post(
        "/oauth/token",
        data={
            "grant_type": DeviceCodeGrant.GRANT_TYPE,
            "device_code": "valid-device",
            "client_id": "invalid",
        },
    )
    resp = json.loads(rv.data)
    assert resp["error"] == "invalid_client"

    client.set_client_metadata(
        {
            "redirect_uris": ["https://client.test/authorized"],
            "scope": "profile",
            "grant_types": ["password"],
            "token_endpoint_auth_method": "none",
        }
    )
    db.session.add(client)
    db.session.commit()

    rv = test_client.post(
        "/oauth/token",
        data={
            "grant_type": DeviceCodeGrant.GRANT_TYPE,
            "device_code": "valid-device",
            "client_id": "client-id",
        },
    )
    resp = json.loads(rv.data)
    assert resp["error"] == "unauthorized_client"


def test_invalid_client(test_client):
    rv = test_client.post(
        "/oauth/token",
        data={
            "grant_type": DeviceCodeGrant.GRANT_TYPE,
            "device_code": "invalid-client",
            "client_id": "invalid",
        },
    )
    resp = json.loads(rv.data)
    assert resp["error"] == "invalid_client"


def test_expired_token(test_client):
    rv = test_client.post(
        "/oauth/token",
        data={
            "grant_type": DeviceCodeGrant.GRANT_TYPE,
            "device_code": "expired-token",
            "client_id": "client-id",
        },
    )
    resp = json.loads(rv.data)
    assert resp["error"] == "expired_token"


def test_denied_by_user(test_client):
    rv = test_client.post(
        "/oauth/token",
        data={
            "grant_type": DeviceCodeGrant.GRANT_TYPE,
            "device_code": "denied-code",
            "client_id": "client-id",
        },
    )
    resp = json.loads(rv.data)
    assert resp["error"] == "access_denied"


def test_authorization_pending(test_client):
    rv = test_client.post(
        "/oauth/token",
        data={
            "grant_type": DeviceCodeGrant.GRANT_TYPE,
            "device_code": "pending-code",
            "client_id": "client-id",
        },
    )
    resp = json.loads(rv.data)
    assert resp["error"] == "authorization_pending"


def test_get_access_token(test_client):
    rv = test_client.post(
        "/oauth/token",
        data={
            "grant_type": DeviceCodeGrant.GRANT_TYPE,
            "device_code": "grant-code",
            "client_id": "client-id",
        },
    )
    resp = json.loads(rv.data)
    assert "access_token" in resp


def test_missing_client_id(test_client):
    rv = test_client.post("/device_authorize", data={"scope": "profile"})
    assert rv.status_code == 401
    resp = json.loads(rv.data)
    assert resp["error"] == "invalid_client"


def test_create_authorization_response(test_client):
    client = Client(
        user_id=1,
        client_id="client",
        client_secret="secret",
    )
    db.session.add(client)
    db.session.commit()
    rv = test_client.post(
        "/device_authorize",
        data={
            "client_id": "client-id",
        },
    )
    assert rv.status_code == 200
    resp = json.loads(rv.data)
    assert "device_code" in resp
    assert "user_code" in resp
    assert resp["verification_uri"] == "https://resource.test/activate"
    assert (
        resp["verification_uri_complete"]
        == "https://resource.test/activate?user_code=" + resp["user_code"]
    )



from authlib.common.encoding import json_dumps


class KeySet:
    """This class represents a JSON Web Key Set."""

    def __init__(self, keys):
        self.keys = keys

    def as_dict(self, is_private=False, **params):
        """Represent this key as a dict of the JSON Web Key Set."""
        return {"keys": [k.as_dict(is_private, **params) for k in self.keys]}

    def as_json(self, is_private=False, **params):
        """Represent this key set as a JSON string."""
        obj = self.as_dict(is_private, **params)
        return json_dumps(obj)

    def find_by_kid(self, kid, **params):
        """Find the key matches the given kid value.

        :param kid: A string of kid
        :return: Key instance
        :raise: ValueError
        """
        # Proposed fix, feel free to do something else but the idea is that we take the only key
        # of the set if no kid is specified
        if kid is None and len(self.keys) == 1:
            return self.keys[0]

        keys = [key for key in self.keys if key.kid == kid]
        if params:
            keys = list(_filter_keys_by_params(keys, **params))

        if keys:
            return keys[0]
        raise ValueError("Key not found")


def _filter_keys_by_params(keys, **params):
    _use = params.get("use")
    _alg = params.get("alg")

    for key in keys:
        designed_use = key.tokens.get("use")
        if designed_use and _use and designed_use != _use:
            continue

        designed_alg = key.tokens.get("alg")
        if designed_alg and _alg and designed_alg != _alg:
            continue

        yield key



from httpx import Request

HTTPX_CLIENT_KWARGS = [
    "headers",
    "cookies",
    "verify",
    "cert",
    "http1",
    "http2",
    "proxy",
    "mounts",
    "timeout",
    "follow_redirects",
    "limits",
    "max_redirects",
    "event_hooks",
    "base_url",
    "transport",
    "trust_env",
    "default_encoding",
]


def extract_client_kwargs(kwargs):
    client_kwargs = {}
    for k in HTTPX_CLIENT_KWARGS:
        if k in kwargs:
            client_kwargs[k] = kwargs.pop(k)
    return client_kwargs


def build_request(url, headers, body, initial_request: Request) -> Request:
    """Make sure that all the data from initial request is passed to the updated object."""
    updated_request = Request(
        method=initial_request.method, url=url, headers=headers, content=body
    )

    if hasattr(initial_request, "extensions"):
        updated_request.extensions = initial_request.extensions

    return updated_request



from collections import OrderedDict

import pytest
from cryptography.hazmat.primitives.keywrap import InvalidUnwrap

from authlib.common.encoding import json_b64encode
from authlib.common.encoding import json_loads
from authlib.common.encoding import to_bytes
from authlib.common.encoding import urlsafe_b64decode
from authlib.common.encoding import urlsafe_b64encode
from authlib.jose import ECKey
from authlib.jose import JsonWebEncryption
from authlib.jose import OKPKey
from authlib.jose.drafts import register_jwe_draft
from authlib.jose.errors import InvalidAlgorithmForMultipleRecipientsMode
from authlib.jose.errors import InvalidEncryptionAlgorithmForECDH1PUWithKeyWrappingError
from authlib.jose.rfc7516.models import JWEHeader

register_jwe_draft(JsonWebEncryption)


def test_ecdh_1pu_key_agreement_computation_appx_a():
    # https://datatracker.ietf.org/doc/html/draft-madden-jose-ecdh-1pu-04#appendix-A
    alice_static_key = {
        "kty": "EC",
        "crv": "P-256",
        "x": "WKn-ZIGevcwGIyyrzFoZNBdaq9_TsqzGl96oc0CWuis",
        "y": "y77t-RvAHRKTsSGdIYUfweuOvwrvDD-Q3Hv5J0fSKbE",
        "d": "Hndv7ZZjs_ke8o9zXYo3iq-Yr8SewI5vrqd0pAvEPqg",
    }
    bob_static_key = {
        "kty": "EC",
        "crv": "P-256",
        "x": "weNJy2HscCSM6AEDTDg04biOvhFhyyWvOHQfeF_PxMQ",
        "y": "e8lnCO-AlStT-NJVX-crhB7QRYhiix03illJOVAOyck",
        "d": "VEmDZpDXXK8p8N0Cndsxs924q6nS1RXFASRl6BfUqdw",
    }
    alice_ephemeral_key = {
        "kty": "EC",
        "crv": "P-256",
        "x": "gI0GAILBdu7T53akrFmMyGcsF3n5dO7MmwNBHKW5SV0",
        "y": "SLW_xSffzlPWrHEVI30DHM_4egVwt3NQqeUD7nMFpps",
        "d": "0_NxaRPUMQoAJt50Gz8YiTr8gRTwyEaCumd-MToTmIo",
    }

    headers = {
        "alg": "ECDH-1PU",
        "enc": "A256GCM",
        "apu": "QWxpY2U",
        "apv": "Qm9i",
        "epk": {
            "kty": "EC",
            "crv": "P-256",
            "x": "gI0GAILBdu7T53akrFmMyGcsF3n5dO7MmwNBHKW5SV0",
            "y": "SLW_xSffzlPWrHEVI30DHM_4egVwt3NQqeUD7nMFpps",
        },
    }

    alg = JsonWebEncryption.ALG_REGISTRY["ECDH-1PU"]
    enc = JsonWebEncryption.ENC_REGISTRY["A256GCM"]

    alice_static_key = alg.prepare_key(alice_static_key)
    bob_static_key = alg.prepare_key(bob_static_key)
    alice_ephemeral_key = alg.prepare_key(alice_ephemeral_key)

    alice_static_pubkey = alice_static_key.get_op_key("wrapKey")
    bob_static_pubkey = bob_static_key.get_op_key("wrapKey")
    alice_ephemeral_pubkey = alice_ephemeral_key.get_op_key("wrapKey")

    # Derived key computation at Alice

    # Step-by-step methods verification
    _shared_key_e_at_alice = alice_ephemeral_key.exchange_shared_key(bob_static_pubkey)
    assert (
        _shared_key_e_at_alice
        == b"\x9e\x56\xd9\x1d\x81\x71\x35\xd3\x72\x83\x42\x83\xbf\x84\x26\x9c"
        + b"\xfb\x31\x6e\xa3\xda\x80\x6a\x48\xf6\xda\xa7\x79\x8c\xfe\x90\xc4"
    )

    _shared_key_s_at_alice = alice_static_key.exchange_shared_key(bob_static_pubkey)
    assert (
        _shared_key_s_at_alice
        == b"\xe3\xca\x34\x74\x38\x4c\x9f\x62\xb3\x0b\xfd\x4c\x68\x8b\x3e\x7d"
        + b"\x41\x10\xa1\xb4\xba\xdc\x3c\xc5\x4e\xf7\xb8\x12\x41\xef\xd5\x0d"
    )

    _shared_key_at_alice = alg.compute_shared_key(
        _shared_key_e_at_alice, _shared_key_s_at_alice
    )
    assert (
        _shared_key_at_alice
        == b"\x9e\x56\xd9\x1d\x81\x71\x35\xd3\x72\x83\x42\x83\xbf\x84\x26\x9c"
        + b"\xfb\x31\x6e\xa3\xda\x80\x6a\x48\xf6\xda\xa7\x79\x8c\xfe\x90\xc4"
        + b"\xe3\xca\x34\x74\x38\x4c\x9f\x62\xb3\x0b\xfd\x4c\x68\x8b\x3e\x7d"
        + b"\x41\x10\xa1\xb4\xba\xdc\x3c\xc5\x4e\xf7\xb8\x12\x41\xef\xd5\x0d"
    )

    _fixed_info_at_alice = alg.compute_fixed_info(headers, enc.key_size, None)
    assert (
        _fixed_info_at_alice
        == b"\x00\x00\x00\x07\x41\x32\x35\x36\x47\x43\x4d\x00\x00\x00\x05\x41"
        + b"\x6c\x69\x63\x65\x00\x00\x00\x03\x42\x6f\x62\x00\x00\x01\x00"
    )

    _dk_at_alice = alg.compute_derived_key(
        _shared_key_at_alice, _fixed_info_at_alice, enc.key_size
    )
    assert (
        _dk_at_alice
        == b"\x6c\xaf\x13\x72\x3d\x14\x85\x0a\xd4\xb4\x2c\xd6\xdd\xe9\x35\xbf"
        + b"\xfd\x2f\xff\x00\xa9\xba\x70\xde\x05\xc2\x03\xa5\xe1\x72\x2c\xa7"
    )
    assert (
        urlsafe_b64encode(_dk_at_alice)
        == b"bK8Tcj0UhQrUtCzW3ek1v_0v_wCpunDeBcIDpeFyLKc"
    )

    # All-in-one method verification
    dk_at_alice = alg.deliver_at_sender(
        alice_static_key,
        alice_ephemeral_key,
        bob_static_pubkey,
        headers,
        enc.key_size,
        None,
    )
    assert (
        urlsafe_b64encode(dk_at_alice) == b"bK8Tcj0UhQrUtCzW3ek1v_0v_wCpunDeBcIDpeFyLKc"
    )

    # Derived key computation at Bob

    # Step-by-step methods verification
    _shared_key_e_at_bob = bob_static_key.exchange_shared_key(alice_ephemeral_pubkey)
    assert _shared_key_e_at_bob == _shared_key_e_at_alice

    _shared_key_s_at_bob = bob_static_key.exchange_shared_key(alice_static_pubkey)
    assert _shared_key_s_at_bob == _shared_key_s_at_alice

    _shared_key_at_bob = alg.compute_shared_key(
        _shared_key_e_at_bob, _shared_key_s_at_bob
    )
    assert _shared_key_at_bob == _shared_key_at_alice

    _fixed_info_at_bob = alg.compute_fixed_info(headers, enc.key_size, None)
    assert _fixed_info_at_bob == _fixed_info_at_alice

    _dk_at_bob = alg.compute_derived_key(
        _shared_key_at_bob, _fixed_info_at_bob, enc.key_size
    )
    assert _dk_at_bob == _dk_at_alice

    # All-in-one method verification
    dk_at_bob = alg.deliver_at_recipient(
        bob_static_key,
        alice_static_pubkey,
        alice_ephemeral_pubkey,
        headers,
        enc.key_size,
        None,
    )
    assert dk_at_bob == dk_at_alice


def test_ecdh_1pu_key_agreement_computation_appx_b():
    # https://datatracker.ietf.org/doc/html/draft-madden-jose-ecdh-1pu-04#appendix-B
    alice_static_key = {
        "kty": "OKP",
        "crv": "X25519",
        "x": "Knbm_BcdQr7WIoz-uqit9M0wbcfEr6y-9UfIZ8QnBD4",
        "d": "i9KuFhSzEBsiv3PKVL5115OCdsqQai5nj_Flzfkw5jU",
    }
    bob_static_key = {
        "kty": "OKP",
        "crv": "X25519",
        "x": "BT7aR0ItXfeDAldeeOlXL_wXqp-j5FltT0vRSG16kRw",
        "d": "1gDirl_r_Y3-qUa3WXHgEXrrEHngWThU3c9zj9A2uBg",
    }
    charlie_static_key = {
        "kty": "OKP",
        "crv": "X25519",
        "x": "q-LsvU772uV_2sPJhfAIq-3vnKNVefNoIlvyvg1hrnE",
        "d": "Jcv8gklhMjC0b-lsk5onBbppWAx5ncNtbM63Jr9xBQE",
    }
    alice_ephemeral_key = {
        "kty": "OKP",
        "crv": "X25519",
        "x": "k9of_cpAajy0poW5gaixXGs9nHkwg1AFqUAFa39dyBc",
        "d": "x8EVZH4Fwk673_mUujnliJoSrLz0zYzzCWp5GUX2fc8",
    }

    protected = OrderedDict(
        {
            "alg": "ECDH-1PU+A128KW",
            "enc": "A256CBC-HS512",
            "apu": "QWxpY2U",
            "apv": "Qm9iIGFuZCBDaGFybGll",
            "epk": OrderedDict(
                {
                    "kty": "OKP",
                    "crv": "X25519",
                    "x": "k9of_cpAajy0poW5gaixXGs9nHkwg1AFqUAFa39dyBc",
                }
            ),
        }
    )

    cek = (
        b"\xff\xfe\xfd\xfc\xfb\xfa\xf9\xf8\xf7\xf6\xf5\xf4\xf3\xf2\xf1\xf0"
        b"\xef\xee\xed\xec\xeb\xea\xe9\xe8\xe7\xe6\xe5\xe4\xe3\xe2\xe1\xe0"
        b"\xdf\xde\xdd\xdc\xdb\xda\xd9\xd8\xd7\xd6\xd5\xd4\xd3\xd2\xd1\xd0"
        b"\xcf\xce\xcd\xcc\xcb\xca\xc9\xc8\xc7\xc6\xc5\xc4\xc3\xc2\xc1\xc0"
    )

    iv = b"\x00\x01\x02\x03\x04\x05\x06\x07\x08\x09\x0a\x0b\x0c\x0d\x0e\x0f"

    payload = b"Three is a magic number."

    alg = JsonWebEncryption.ALG_REGISTRY["ECDH-1PU+A128KW"]
    enc = JsonWebEncryption.ENC_REGISTRY["A256CBC-HS512"]

    alice_static_key = OKPKey.import_key(alice_static_key)
    bob_static_key = OKPKey.import_key(bob_static_key)
    charlie_static_key = OKPKey.import_key(charlie_static_key)
    alice_ephemeral_key = OKPKey.import_key(alice_ephemeral_key)

    alice_static_pubkey = alice_static_key.get_op_key("wrapKey")
    bob_static_pubkey = bob_static_key.get_op_key("wrapKey")
    charlie_static_pubkey = charlie_static_key.get_op_key("wrapKey")
    alice_ephemeral_pubkey = alice_ephemeral_key.get_op_key("wrapKey")

    protected_segment = json_b64encode(protected)
    aad = to_bytes(protected_segment, "ascii")

    ciphertext, tag = enc.encrypt(payload, aad, iv, cek)
    assert (
        urlsafe_b64encode(ciphertext) == b"Az2IWsISEMDJvyc5XRL-3-d-RgNBOGolCsxFFoUXFYw"
    )
    assert urlsafe_b64encode(tag) == b"HLb4fTlm8spGmij3RyOs2gJ4DpHM4hhVRwdF_hGb3WQ"

    # Derived key computation at Alice for Bob

    # Step-by-step methods verification
    _shared_key_e_at_alice_for_bob = alice_ephemeral_key.exchange_shared_key(
        bob_static_pubkey
    )
    assert (
        _shared_key_e_at_alice_for_bob
        == b"\x32\x81\x08\x96\xe0\xfe\x4d\x57\x0e\xd1\xac\xfc\xed\xf6\x71\x17"
        + b"\xdc\x19\x4e\xd5\xda\xac\x21\xd8\xff\x7a\xf3\x24\x46\x94\x89\x7f"
    )

    _shared_key_s_at_alice_for_bob = alice_static_key.exchange_shared_key(
        bob_static_pubkey
    )
    assert (
        _shared_key_s_at_alice_for_bob
        == b"\x21\x57\x61\x2c\x90\x48\xed\xfa\xe7\x7c\xb2\xe4\x23\x71\x40\x60"
        + b"\x59\x67\xc0\x5c\x7f\x77\xa4\x8e\xea\xf2\xcf\x29\xa5\x73\x7c\x4a"
    )

    _shared_key_at_alice_for_bob = alg.compute_shared_key(
        _shared_key_e_at_alice_for_bob, _shared_key_s_at_alice_for_bob
    )
    assert (
        _shared_key_at_alice_for_bob
        == b"\x32\x81\x08\x96\xe0\xfe\x4d\x57\x0e\xd1\xac\xfc\xed\xf6\x71\x17"
        + b"\xdc\x19\x4e\xd5\xda\xac\x21\xd8\xff\x7a\xf3\x24\x46\x94\x89\x7f"
        + b"\x21\x57\x61\x2c\x90\x48\xed\xfa\xe7\x7c\xb2\xe4\x23\x71\x40\x60"
        + b"\x59\x67\xc0\x5c\x7f\x77\xa4\x8e\xea\xf2\xcf\x29\xa5\x73\x7c\x4a"
    )

    _fixed_info_at_alice_for_bob = alg.compute_fixed_info(protected, alg.key_size, tag)
    assert (
        _fixed_info_at_alice_for_bob
        == b"\x00\x00\x00\x0f\x45\x43\x44\x48\x2d\x31\x50\x55\x2b\x41\x31\x32"
        + b"\x38\x4b\x57\x00\x00\x00\x05\x41\x6c\x69\x63\x65\x00\x00\x00\x0f"
        + b"\x42\x6f\x62\x20\x61\x6e\x64\x20\x43\x68\x61\x72\x6c\x69\x65\x00"
        + b"\x00\x00\x80\x00\x00\x00\x20\x1c\xb6\xf8\x7d\x39\x66\xf2\xca\x46"
        + b"\x9a\x28\xf7\x47\x23\xac\xda\x02\x78\x0e\x91\xcc\xe2\x18\x55\x47"
        + b"\x07\x45\xfe\x11\x9b\xdd\x64"
    )

    _dk_at_alice_for_bob = alg.compute_derived_key(
        _shared_key_at_alice_for_bob, _fixed_info_at_alice_for_bob, alg.key_size
    )
    assert (
        _dk_at_alice_for_bob
        == b"\xdf\x4c\x37\xa0\x66\x83\x06\xa1\x1e\x3d\x6b\x00\x74\xb5\xd8\xdf"
    )

    # All-in-one method verification
    dk_at_alice_for_bob = alg.deliver_at_sender(
        alice_static_key,
        alice_ephemeral_key,
        bob_static_pubkey,
        protected,
        alg.key_size,
        tag,
    )
    assert (
        dk_at_alice_for_bob
        == b"\xdf\x4c\x37\xa0\x66\x83\x06\xa1\x1e\x3d\x6b\x00\x74\xb5\xd8\xdf"
    )

    kek_at_alice_for_bob = alg.aeskw.prepare_key(dk_at_alice_for_bob)
    wrapped_for_bob = alg.aeskw.wrap_cek(cek, kek_at_alice_for_bob)
    ek_for_bob = wrapped_for_bob["ek"]
    assert (
        urlsafe_b64encode(ek_for_bob)
        == b"pOMVA9_PtoRe7xXW1139NzzN1UhiFoio8lGto9cf0t8PyU-sjNXH8-LIRLycq8CHJQbDwvQeU1cSl55cQ0hGezJu2N9IY0QN"
    )

    # Derived key computation at Alice for Charlie

    # Step-by-step methods verification
    _shared_key_e_at_alice_for_charlie = alice_ephemeral_key.exchange_shared_key(
        charlie_static_pubkey
    )
    assert (
        _shared_key_e_at_alice_for_charlie
        == b"\x89\xdc\xfe\x4c\x37\xc1\xdc\x02\x71\xf3\x46\xb5\xb3\xb1\x9c\x3b"
        + b"\x70\x5c\xa2\xa7\x2f\x9a\x23\x77\x85\xc3\x44\x06\xfc\xb7\x5f\x10"
    )

    _shared_key_s_at_alice_for_charlie = alice_static_key.exchange_shared_key(
        charlie_static_pubkey
    )
    assert (
        _shared_key_s_at_alice_for_charlie
        == b"\x78\xfe\x63\xfc\x66\x1c\xf8\xd1\x8f\x92\xa8\x42\x2a\x64\x18\xe4"
        + b"\xed\x5e\x20\xa9\x16\x81\x85\xfd\xee\xdc\xa1\xc3\xd8\xe6\xa6\x1c"
    )

    _shared_key_at_alice_for_charlie = alg.compute_shared_key(
        _shared_key_e_at_alice_for_charlie, _shared_key_s_at_alice_for_charlie
    )
    assert (
        _shared_key_at_alice_for_charlie
        == b"\x89\xdc\xfe\x4c\x37\xc1\xdc\x02\x71\xf3\x46\xb5\xb3\xb1\x9c\x3b"
        + b"\x70\x5c\xa2\xa7\x2f\x9a\x23\x77\x85\xc3\x44\x06\xfc\xb7\x5f\x10"
        + b"\x78\xfe\x63\xfc\x66\x1c\xf8\xd1\x8f\x92\xa8\x42\x2a\x64\x18\xe4"
        + b"\xed\x5e\x20\xa9\x16\x81\x85\xfd\xee\xdc\xa1\xc3\xd8\xe6\xa6\x1c"
    )

    _fixed_info_at_alice_for_charlie = alg.compute_fixed_info(
        protected, alg.key_size, tag
    )
    assert _fixed_info_at_alice_for_charlie == _fixed_info_at_alice_for_bob

    _dk_at_alice_for_charlie = alg.compute_derived_key(
        _shared_key_at_alice_for_charlie,
        _fixed_info_at_alice_for_charlie,
        alg.key_size,
    )
    assert (
        _dk_at_alice_for_charlie
        == b"\x57\xd8\x12\x6f\x1b\x7e\xc4\xcc\xb0\x58\x4d\xac\x03\xcb\x27\xcc"
    )

    # All-in-one method verification
    dk_at_alice_for_charlie = alg.deliver_at_sender(
        alice_static_key,
        alice_ephemeral_key,
        charlie_static_pubkey,
        protected,
        alg.key_size,
        tag,
    )
    assert (
        dk_at_alice_for_charlie
        == b"\x57\xd8\x12\x6f\x1b\x7e\xc4\xcc\xb0\x58\x4d\xac\x03\xcb\x27\xcc"
    )

    kek_at_alice_for_charlie = alg.aeskw.prepare_key(dk_at_alice_for_charlie)
    wrapped_for_charlie = alg.aeskw.wrap_cek(cek, kek_at_alice_for_charlie)
    ek_for_charlie = wrapped_for_charlie["ek"]
    assert (
        urlsafe_b64encode(ek_for_charlie)
        == b"56GVudgRLIMEElQ7DpXsijJVRSWUSDNdbWkdV3g0GUNq6hcT_GkxwnxlPIWrTXCqRpVKQC8fe4z3PQ2YH2afvjQ28aiCTWFE"
    )

    # Derived key computation at Bob for Alice

    # Step-by-step methods verification
    _shared_key_e_at_bob_for_alice = bob_static_key.exchange_shared_key(
        alice_ephemeral_pubkey
    )
    assert _shared_key_e_at_bob_for_alice == _shared_key_e_at_alice_for_bob

    _shared_key_s_at_bob_for_alice = bob_static_key.exchange_shared_key(
        alice_static_pubkey
    )
    assert _shared_key_s_at_bob_for_alice == _shared_key_s_at_alice_for_bob

    _shared_key_at_bob_for_alice = alg.compute_shared_key(
        _shared_key_e_at_bob_for_alice, _shared_key_s_at_bob_for_alice
    )
    assert _shared_key_at_bob_for_alice == _shared_key_at_alice_for_bob

    _fixed_info_at_bob_for_alice = alg.compute_fixed_info(protected, alg.key_size, tag)
    assert _fixed_info_at_bob_for_alice == _fixed_info_at_alice_for_bob

    _dk_at_bob_for_alice = alg.compute_derived_key(
        _shared_key_at_bob_for_alice, _fixed_info_at_bob_for_alice, alg.key_size
    )
    assert _dk_at_bob_for_alice == _dk_at_alice_for_bob

    # All-in-one method verification
    dk_at_bob_for_alice = alg.deliver_at_recipient(
        bob_static_key,
        alice_static_pubkey,
        alice_ephemeral_pubkey,
        protected,
        alg.key_size,
        tag,
    )
    assert dk_at_bob_for_alice == dk_at_alice_for_bob

    kek_at_bob_for_alice = alg.aeskw.prepare_key(dk_at_bob_for_alice)
    cek_unwrapped_by_bob = alg.aeskw.unwrap(
        enc, ek_for_bob, protected, kek_at_bob_for_alice
    )
    assert cek_unwrapped_by_bob == cek

    payload_decrypted_by_bob = enc.decrypt(
        ciphertext, aad, iv, tag, cek_unwrapped_by_bob
    )
    assert payload_decrypted_by_bob == payload

    # Derived key computation at Charlie for Alice

    # Step-by-step methods verification
    _shared_key_e_at_charlie_for_alice = charlie_static_key.exchange_shared_key(
        alice_ephemeral_pubkey
    )
    assert _shared_key_e_at_charlie_for_alice == _shared_key_e_at_alice_for_charlie

    _shared_key_s_at_charlie_for_alice = charlie_static_key.exchange_shared_key(
        alice_static_pubkey
    )
    assert _shared_key_s_at_charlie_for_alice == _shared_key_s_at_alice_for_charlie

    _shared_key_at_charlie_for_alice = alg.compute_shared_key(
        _shared_key_e_at_charlie_for_alice, _shared_key_s_at_charlie_for_alice
    )
    assert _shared_key_at_charlie_for_alice == _shared_key_at_alice_for_charlie

    _fixed_info_at_charlie_for_alice = alg.compute_fixed_info(
        protected, alg.key_size, tag
    )
    assert _fixed_info_at_charlie_for_alice == _fixed_info_at_alice_for_charlie

    _dk_at_charlie_for_alice = alg.compute_derived_key(
        _shared_key_at_charlie_for_alice,
        _fixed_info_at_charlie_for_alice,
        alg.key_size,
    )
    assert _dk_at_charlie_for_alice == _dk_at_alice_for_charlie

    # All-in-one method verification
    dk_at_charlie_for_alice = alg.deliver_at_recipient(
        charlie_static_key,
        alice_static_pubkey,
        alice_ephemeral_pubkey,
        protected,
        alg.key_size,
        tag,
    )
    assert dk_at_charlie_for_alice == dk_at_alice_for_charlie

    kek_at_charlie_for_alice = alg.aeskw.prepare_key(dk_at_charlie_for_alice)
    cek_unwrapped_by_charlie = alg.aeskw.unwrap(
        enc, ek_for_charlie, protected, kek_at_charlie_for_alice
    )
    assert cek_unwrapped_by_charlie == cek

    payload_decrypted_by_charlie = enc.decrypt(
        ciphertext, aad, iv, tag, cek_unwrapped_by_charlie
    )
    assert payload_decrypted_by_charlie == payload


def test_ecdh_1pu_jwe_in_direct_key_agreement_mode():
    jwe = JsonWebEncryption()
    alice_key = {
        "kty": "EC",
        "crv": "P-256",
        "x": "WKn-ZIGevcwGIyyrzFoZNBdaq9_TsqzGl96oc0CWuis",
        "y": "y77t-RvAHRKTsSGdIYUfweuOvwrvDD-Q3Hv5J0fSKbE",
        "d": "Hndv7ZZjs_ke8o9zXYo3iq-Yr8SewI5vrqd0pAvEPqg",
    }
    bob_key = {
        "kty": "EC",
        "crv": "P-256",
        "x": "weNJy2HscCSM6AEDTDg04biOvhFhyyWvOHQfeF_PxMQ",
        "y": "e8lnCO-AlStT-NJVX-crhB7QRYhiix03illJOVAOyck",
        "d": "VEmDZpDXXK8p8N0Cndsxs924q6nS1RXFASRl6BfUqdw",
    }

    for enc in [
        "A128CBC-HS256",
        "A192CBC-HS384",
        "A256CBC-HS512",
        "A128GCM",
        "A192GCM",
        "A256GCM",
    ]:
        protected = {"alg": "ECDH-1PU", "enc": enc}
        data = jwe.serialize_compact(protected, b"hello", bob_key, sender_key=alice_key)
        rv = jwe.deserialize_compact(data, bob_key, sender_key=alice_key)
        assert rv["payload"] == b"hello"


def test_ecdh_1pu_jwe_json_serialization_single_recipient_in_direct_key_agreement_mode():
    jwe = JsonWebEncryption()
    alice_key = OKPKey.generate_key("X25519", is_private=True)
    bob_key = OKPKey.generate_key("X25519", is_private=True)

    protected = {"alg": "ECDH-1PU", "enc": "A128GCM"}
    header_obj = {"protected": protected}
    data = jwe.serialize_json(header_obj, b"hello", bob_key, sender_key=alice_key)
    rv = jwe.deserialize_json(data, bob_key, sender_key=alice_key)
    assert rv["payload"] == b"hello"


def test_ecdh_1pu_jwe_in_key_agreement_with_key_wrapping_mode():
    jwe = JsonWebEncryption()
    alice_key = {
        "kty": "EC",
        "crv": "P-256",
        "x": "WKn-ZIGevcwGIyyrzFoZNBdaq9_TsqzGl96oc0CWuis",
        "y": "y77t-RvAHRKTsSGdIYUfweuOvwrvDD-Q3Hv5J0fSKbE",
        "d": "Hndv7ZZjs_ke8o9zXYo3iq-Yr8SewI5vrqd0pAvEPqg",
    }
    bob_key = {
        "kty": "EC",
        "crv": "P-256",
        "x": "weNJy2HscCSM6AEDTDg04biOvhFhyyWvOHQfeF_PxMQ",
        "y": "e8lnCO-AlStT-NJVX-crhB7QRYhiix03illJOVAOyck",
        "d": "VEmDZpDXXK8p8N0Cndsxs924q6nS1RXFASRl6BfUqdw",
    }

    for alg in [
        "ECDH-1PU+A128KW",
        "ECDH-1PU+A192KW",
        "ECDH-1PU+A256KW",
    ]:
        for enc in [
            "A128CBC-HS256",
            "A192CBC-HS384",
            "A256CBC-HS512",
        ]:
            protected = {"alg": alg, "enc": enc}
            data = jwe.serialize_compact(
                protected, b"hello", bob_key, sender_key=alice_key
            )
            rv = jwe.deserialize_compact(data, bob_key, sender_key=alice_key)
            assert rv["payload"] == b"hello"


def test_ecdh_1pu_jwe_with_compact_serialization_ignores_kid_provided_separately_on_decryption():
    jwe = JsonWebEncryption()

    alice_key = {
        "kty": "EC",
        "crv": "P-256",
        "x": "WKn-ZIGevcwGIyyrzFoZNBdaq9_TsqzGl96oc0CWuis",
        "y": "y77t-RvAHRKTsSGdIYUfweuOvwrvDD-Q3Hv5J0fSKbE",
        "d": "Hndv7ZZjs_ke8o9zXYo3iq-Yr8SewI5vrqd0pAvEPqg",
    }

    bob_kid = "Bob's key"
    bob_key = {
        "kty": "EC",
        "crv": "P-256",
        "x": "weNJy2HscCSM6AEDTDg04biOvhFhyyWvOHQfeF_PxMQ",
        "y": "e8lnCO-AlStT-NJVX-crhB7QRYhiix03illJOVAOyck",
        "d": "VEmDZpDXXK8p8N0Cndsxs924q6nS1RXFASRl6BfUqdw",
    }

    for alg in [
        "ECDH-1PU+A128KW",
        "ECDH-1PU+A192KW",
        "ECDH-1PU+A256KW",
    ]:
        for enc in [
            "A128CBC-HS256",
            "A192CBC-HS384",
            "A256CBC-HS512",
        ]:
            protected = {"alg": alg, "enc": enc}
            data = jwe.serialize_compact(
                protected, b"hello", bob_key, sender_key=alice_key
            )
            rv = jwe.deserialize_compact(data, (bob_kid, bob_key), sender_key=alice_key)
            assert rv["payload"] == b"hello"


def test_ecdh_1pu_jwe_with_okp_keys_in_direct_key_agreement_mode():
    jwe = JsonWebEncryption()
    alice_key = OKPKey.generate_key("X25519", is_private=True)
    bob_key = OKPKey.generate_key("X25519", is_private=True)

    for enc in [
        "A128CBC-HS256",
        "A192CBC-HS384",
        "A256CBC-HS512",
        "A128GCM",
        "A192GCM",
        "A256GCM",
    ]:
        protected = {"alg": "ECDH-1PU", "enc": enc}
        data = jwe.serialize_compact(protected, b"hello", bob_key, sender_key=alice_key)
        rv = jwe.deserialize_compact(data, bob_key, sender_key=alice_key)
        assert rv["payload"] == b"hello"


def test_ecdh_1pu_jwe_with_okp_keys_in_key_agreement_with_key_wrapping_mode():
    jwe = JsonWebEncryption()
    alice_key = OKPKey.generate_key("X25519", is_private=True)
    bob_key = OKPKey.generate_key("X25519", is_private=True)

    for alg in [
        "ECDH-1PU+A128KW",
        "ECDH-1PU+A192KW",
        "ECDH-1PU+A256KW",
    ]:
        for enc in [
            "A128CBC-HS256",
            "A192CBC-HS384",
            "A256CBC-HS512",
        ]:
            protected = {"alg": alg, "enc": enc}
            data = jwe.serialize_compact(
                protected, b"hello", bob_key, sender_key=alice_key
            )
            rv = jwe.deserialize_compact(data, bob_key, sender_key=alice_key)
            assert rv["payload"] == b"hello"


def test_ecdh_1pu_encryption_with_json_serialization():
    jwe = JsonWebEncryption()

    alice_key = OKPKey.import_key(
        {
            "kty": "OKP",
            "crv": "X25519",
            "x": "Knbm_BcdQr7WIoz-uqit9M0wbcfEr6y-9UfIZ8QnBD4",
            "d": "i9KuFhSzEBsiv3PKVL5115OCdsqQai5nj_Flzfkw5jU",
        }
    )
    bob_key = OKPKey.import_key(
        {
            "kty": "OKP",
            "crv": "X25519",
            "x": "BT7aR0ItXfeDAldeeOlXL_wXqp-j5FltT0vRSG16kRw",
            "d": "1gDirl_r_Y3-qUa3WXHgEXrrEHngWThU3c9zj9A2uBg",
        }
    )
    charlie_key = OKPKey.import_key(
        {
            "kty": "OKP",
            "crv": "X25519",
            "x": "q-LsvU772uV_2sPJhfAIq-3vnKNVefNoIlvyvg1hrnE",
            "d": "Jcv8gklhMjC0b-lsk5onBbppWAx5ncNtbM63Jr9xBQE",
        }
    )

    protected = {
        "alg": "ECDH-1PU+A128KW",
        "enc": "A256CBC-HS512",
        "apu": "QWxpY2U",
        "apv": "Qm9iIGFuZCBDaGFybGll",
    }

    unprotected = {"jku": "https://provider.test/jwks"}

    recipients = [
        {"header": {"kid": "bob-key-2"}},
        {"header": {"kid": "2021-05-06"}},
    ]

    jwe_aad = b"Authenticate me too."

    header_obj = {
        "protected": protected,
        "unprotected": unprotected,
        "recipients": recipients,
        "aad": jwe_aad,
    }

    payload = b"Three is a magic number."

    data = jwe.serialize_json(
        header_obj, payload, [bob_key, charlie_key], sender_key=alice_key
    )

    assert data.keys() == {
        "protected",
        "unprotected",
        "recipients",
        "aad",
        "iv",
        "ciphertext",
        "tag",
    }

    decoded_protected = json_loads(
        urlsafe_b64decode(to_bytes(data["protected"])).decode("utf-8")
    )
    assert decoded_protected.keys() == protected.keys() | {"epk"}
    assert {
        k: decoded_protected[k] for k in decoded_protected.keys() - {"epk"}
    } == protected

    assert data["unprotected"] == unprotected

    assert len(data["recipients"]) == len(recipients)
    for i in range(len(data["recipients"])):
        assert data["recipients"][i].keys() == {"header", "encrypted_key"}
        assert data["recipients"][i]["header"] == recipients[i]["header"]

    assert urlsafe_b64decode(to_bytes(data["aad"])) == jwe_aad

    iv = urlsafe_b64decode(to_bytes(data["iv"]))
    ciphertext = urlsafe_b64decode(to_bytes(data["ciphertext"]))
    tag = urlsafe_b64decode(to_bytes(data["tag"]))

    alg = JsonWebEncryption.ALG_REGISTRY[protected["alg"]]
    enc = JsonWebEncryption.ENC_REGISTRY[protected["enc"]]

    aad = to_bytes(data["protected"]) + b"." + to_bytes(data["aad"])
    aad = to_bytes(aad, "ascii")

    ek_for_bob = urlsafe_b64decode(to_bytes(data["recipients"][0]["encrypted_key"]))
    header_for_bob = JWEHeader(
        decoded_protected, data["unprotected"], data["recipients"][0]["header"]
    )
    cek_at_bob = alg.unwrap(
        enc, ek_for_bob, header_for_bob, bob_key, sender_key=alice_key, tag=tag
    )
    payload_at_bob = enc.decrypt(ciphertext, aad, iv, tag, cek_at_bob)

    assert payload_at_bob == payload

    ek_for_charlie = urlsafe_b64decode(to_bytes(data["recipients"][1]["encrypted_key"]))
    header_for_charlie = JWEHeader(
        decoded_protected, data["unprotected"], data["recipients"][1]["header"]
    )
    cek_at_charlie = alg.unwrap(
        enc,
        ek_for_charlie,
        header_for_charlie,
        charlie_key,
        sender_key=alice_key,
        tag=tag,
    )
    payload_at_charlie = enc.decrypt(ciphertext, aad, iv, tag, cek_at_charlie)

    assert cek_at_charlie == cek_at_bob
    assert payload_at_charlie == payload


def test_ecdh_1pu_decryption_with_json_serialization():
    jwe = JsonWebEncryption()

    alice_key = OKPKey.import_key(
        {
            "kty": "OKP",
            "crv": "X25519",
            "x": "Knbm_BcdQr7WIoz-uqit9M0wbcfEr6y-9UfIZ8QnBD4",
            "d": "i9KuFhSzEBsiv3PKVL5115OCdsqQai5nj_Flzfkw5jU",
        }
    )
    bob_key = OKPKey.import_key(
        {
            "kty": "OKP",
            "crv": "X25519",
            "x": "BT7aR0ItXfeDAldeeOlXL_wXqp-j5FltT0vRSG16kRw",
            "d": "1gDirl_r_Y3-qUa3WXHgEXrrEHngWThU3c9zj9A2uBg",
        }
    )
    charlie_key = OKPKey.import_key(
        {
            "kty": "OKP",
            "crv": "X25519",
            "x": "q-LsvU772uV_2sPJhfAIq-3vnKNVefNoIlvyvg1hrnE",
            "d": "Jcv8gklhMjC0b-lsk5onBbppWAx5ncNtbM63Jr9xBQE",
        }
    )

    data = {
        "protected": "eyJhbGciOiJFQ0RILTFQVStBMTI4S1ciLCJlbmMiOiJBMjU2Q0JDLUhTNTEyIiwiYXB1Ijoi"
        + "UVd4cFkyVSIsImFwdiI6IlFtOWlJR0Z1WkNCRGFHRnliR2xsIiwiZXBrIjp7Imt0eSI6Ik9L"
        + "UCIsImNydiI6IlgyNTUxOSIsIngiOiJrOW9mX2NwQWFqeTBwb1c1Z2FpeFhHczluSGt3ZzFB"
        + "RnFVQUZhMzlkeUJjIn19",
        "unprotected": {"jku": "https://provider.test/jwks"},
        "recipients": [
            {
                "header": {"kid": "bob-key-2"},
                "encrypted_key": "pOMVA9_PtoRe7xXW1139NzzN1UhiFoio8lGto9cf0t8PyU-sjNXH8-LIRLycq8CHJQbDwvQ"
                + "eU1cSl55cQ0hGezJu2N9IY0QN",
            },
            {
                "header": {"kid": "2021-05-06"},
                "encrypted_key": "56GVudgRLIMEElQ7DpXsijJVRSWUSDNdbWkdV3g0GUNq6hcT_GkxwnxlPIWrTXCqRpVKQC8"
                + "fe4z3PQ2YH2afvjQ28aiCTWFE",
            },
        ],
        "iv": "AAECAwQFBgcICQoLDA0ODw",
        "ciphertext": "Az2IWsISEMDJvyc5XRL-3-d-RgNBOGolCsxFFoUXFYw",
        "tag": "HLb4fTlm8spGmij3RyOs2gJ4DpHM4hhVRwdF_hGb3WQ",
    }

    rv_at_bob = jwe.deserialize_json(data, bob_key, sender_key=alice_key)

    assert rv_at_bob.keys() == {"header", "payload"}

    assert rv_at_bob["header"].keys() == {"protected", "unprotected", "recipients"}

    assert rv_at_bob["header"]["protected"] == {
        "alg": "ECDH-1PU+A128KW",
        "enc": "A256CBC-HS512",
        "apu": "QWxpY2U",
        "apv": "Qm9iIGFuZCBDaGFybGll",
        "epk": {
            "kty": "OKP",
            "crv": "X25519",
            "x": "k9of_cpAajy0poW5gaixXGs9nHkwg1AFqUAFa39dyBc",
        },
    }

    assert rv_at_bob["header"]["unprotected"] == {"jku": "https://provider.test/jwks"}

    assert rv_at_bob["header"]["recipients"] == [
        {"header": {"kid": "bob-key-2"}},
        {"header": {"kid": "2021-05-06"}},
    ]

    assert rv_at_bob["payload"] == b"Three is a magic number."

    rv_at_charlie = jwe.deserialize_json(data, charlie_key, sender_key=alice_key)

    assert rv_at_charlie.keys() == {"header", "payload"}

    assert rv_at_charlie["header"].keys() == {
        "protected",
        "unprotected",
        "recipients",
    }

    assert rv_at_charlie["header"]["protected"] == {
        "alg": "ECDH-1PU+A128KW",
        "enc": "A256CBC-HS512",
        "apu": "QWxpY2U",
        "apv": "Qm9iIGFuZCBDaGFybGll",
        "epk": {
            "kty": "OKP",
            "crv": "X25519",
            "x": "k9of_cpAajy0poW5gaixXGs9nHkwg1AFqUAFa39dyBc",
        },
    }

    assert rv_at_charlie["header"]["unprotected"] == {
        "jku": "https://provider.test/jwks"
    }

    assert rv_at_charlie["header"]["recipients"] == [
        {"header": {"kid": "bob-key-2"}},
        {"header": {"kid": "2021-05-06"}},
    ]

    assert rv_at_charlie["payload"] == b"Three is a magic number."


def test_ecdh_1pu_jwe_with_json_serialization_when_kid_is_not_specified():
    jwe = JsonWebEncryption()

    alice_key = OKPKey.import_key(
        {
            "kty": "OKP",
            "crv": "X25519",
            "x": "Knbm_BcdQr7WIoz-uqit9M0wbcfEr6y-9UfIZ8QnBD4",
            "d": "i9KuFhSzEBsiv3PKVL5115OCdsqQai5nj_Flzfkw5jU",
        }
    )
    bob_key = OKPKey.import_key(
        {
            "kty": "OKP",
            "crv": "X25519",
            "x": "BT7aR0ItXfeDAldeeOlXL_wXqp-j5FltT0vRSG16kRw",
            "d": "1gDirl_r_Y3-qUa3WXHgEXrrEHngWThU3c9zj9A2uBg",
        }
    )
    charlie_key = OKPKey.import_key(
        {
            "kty": "OKP",
            "crv": "X25519",
            "x": "q-LsvU772uV_2sPJhfAIq-3vnKNVefNoIlvyvg1hrnE",
            "d": "Jcv8gklhMjC0b-lsk5onBbppWAx5ncNtbM63Jr9xBQE",
        }
    )

    protected = {
        "alg": "ECDH-1PU+A128KW",
        "enc": "A256CBC-HS512",
        "apu": "QWxpY2U",
        "apv": "Qm9iIGFuZCBDaGFybGll",
    }

    unprotected = {"jku": "https://provider.test/jwks"}

    recipients = [
        {"header": {"kid": "bob-key-2"}},
        {"header": {"kid": "2021-05-06"}},
    ]

    jwe_aad = b"Authenticate me too."

    header_obj = {
        "protected": protected,
        "unprotected": unprotected,
        "recipients": recipients,
        "aad": jwe_aad,
    }

    payload = b"Three is a magic number."

    data = jwe.serialize_json(
        header_obj, payload, [bob_key, charlie_key], sender_key=alice_key
    )

    rv_at_bob = jwe.deserialize_json(data, bob_key, sender_key=alice_key)

    assert rv_at_bob["header"]["protected"].keys() == protected.keys() | {"epk"}
    assert {
        k: rv_at_bob["header"]["protected"][k]
        for k in rv_at_bob["header"]["protected"].keys() - {"epk"}
    } == protected
    assert rv_at_bob["header"]["unprotected"] == unprotected
    assert rv_at_bob["header"]["recipients"] == recipients
    assert rv_at_bob["header"]["aad"] == jwe_aad
    assert rv_at_bob["payload"] == payload

    rv_at_charlie = jwe.deserialize_json(data, charlie_key, sender_key=alice_key)

    assert rv_at_charlie["header"]["protected"].keys() == protected.keys() | {"epk"}
    assert {
        k: rv_at_charlie["header"]["protected"][k]
        for k in rv_at_charlie["header"]["protected"].keys() - {"epk"}
    } == protected
    assert rv_at_charlie["header"]["unprotected"] == unprotected
    assert rv_at_charlie["header"]["recipients"] == recipients
    assert rv_at_charlie["header"]["aad"] == jwe_aad
    assert rv_at_charlie["payload"] == payload


def test_ecdh_1pu_jwe_with_json_serialization_when_kid_is_specified():
    jwe = JsonWebEncryption()

    alice_key = OKPKey.import_key(
        {
            "kty": "OKP",
            "crv": "X25519",
            "kid": "alice-key",
            "x": "Knbm_BcdQr7WIoz-uqit9M0wbcfEr6y-9UfIZ8QnBD4",
            "d": "i9KuFhSzEBsiv3PKVL5115OCdsqQai5nj_Flzfkw5jU",
        }
    )
    bob_key = OKPKey.import_key(
        {
            "kty": "OKP",
            "crv": "X25519",
            "kid": "bob-key-2",
            "x": "BT7aR0ItXfeDAldeeOlXL_wXqp-j5FltT0vRSG16kRw",
            "d": "1gDirl_r_Y3-qUa3WXHgEXrrEHngWThU3c9zj9A2uBg",
        }
    )
    charlie_key = OKPKey.import_key(
        {
            "kty": "OKP",
            "crv": "X25519",
            "kid": "2021-05-06",
            "x": "q-LsvU772uV_2sPJhfAIq-3vnKNVefNoIlvyvg1hrnE",
            "d": "Jcv8gklhMjC0b-lsk5onBbppWAx5ncNtbM63Jr9xBQE",
        }
    )

    protected = {
        "alg": "ECDH-1PU+A128KW",
        "enc": "A256CBC-HS512",
        "apu": "QWxpY2U",
        "apv": "Qm9iIGFuZCBDaGFybGll",
    }

    unprotected = {"jku": "https://provider.test/jwks"}

    recipients = [
        {"header": {"kid": "bob-key-2"}},
        {"header": {"kid": "2021-05-06"}},
    ]

    jwe_aad = b"Authenticate me too."

    header_obj = {
        "protected": protected,
        "unprotected": unprotected,
        "recipients": recipients,
        "aad": jwe_aad,
    }

    payload = b"Three is a magic number."

    data = jwe.serialize_json(
        header_obj, payload, [bob_key, charlie_key], sender_key=alice_key
    )

    rv_at_bob = jwe.deserialize_json(data, bob_key, sender_key=alice_key)

    assert rv_at_bob["header"]["protected"].keys() == protected.keys() | {"epk"}
    assert {
        k: rv_at_bob["header"]["protected"][k]
        for k in rv_at_bob["header"]["protected"].keys() - {"epk"}
    } == protected
    assert rv_at_bob["header"]["unprotected"] == unprotected
    assert rv_at_bob["header"]["recipients"] == recipients
    assert rv_at_bob["header"]["aad"] == jwe_aad
    assert rv_at_bob["payload"] == payload

    rv_at_charlie = jwe.deserialize_json(data, charlie_key, sender_key=alice_key)

    assert rv_at_charlie["header"]["protected"].keys() == protected.keys() | {"epk"}
    assert {
        k: rv_at_charlie["header"]["protected"][k]
        for k in rv_at_charlie["header"]["protected"].keys() - {"epk"}
    } == protected
    assert rv_at_charlie["header"]["unprotected"] == unprotected
    assert rv_at_charlie["header"]["recipients"] == recipients
    assert rv_at_charlie["header"]["aad"] == jwe_aad
    assert rv_at_charlie["payload"] == payload


def test_ecdh_1pu_jwe_with_json_serialization_when_kid_is_provided_separately_on_decryption():
    jwe = JsonWebEncryption()

    alice_key = OKPKey.import_key(
        {
            "kty": "OKP",
            "crv": "X25519",
            "kid": "WjKgJV7VRw3hmgU6--4v15c0Aewbcvat1BsRFTIqa5Q",
            "x": "Knbm_BcdQr7WIoz-uqit9M0wbcfEr6y-9UfIZ8QnBD4",
            "d": "i9KuFhSzEBsiv3PKVL5115OCdsqQai5nj_Flzfkw5jU",
        }
    )

    bob_kid = "did:example:123#_Qq0UL2Fq651Q0Fjd6TvnYE-faHiOpRlPVQcY_-tA4A"
    bob_key = OKPKey.import_key(
        {
            "kty": "OKP",
            "crv": "X25519",
            "kid": "_Qq0UL2Fq651Q0Fjd6TvnYE-faHiOpRlPVQcY_-tA4A",
            "x": "BT7aR0ItXfeDAldeeOlXL_wXqp-j5FltT0vRSG16kRw",
            "d": "1gDirl_r_Y3-qUa3WXHgEXrrEHngWThU3c9zj9A2uBg",
        }
    )

    charlie_kid = "did:example:123#_TKzHv2jFIyvdTGF1Dsgwngfdg3SH6TpDv0Ta1aOEkw"
    charlie_key = OKPKey.import_key(
        {
            "kty": "OKP",
            "crv": "X25519",
            "kid": "_TKzHv2jFIyvdTGF1Dsgwngfdg3SH6TpDv0Ta1aOEkw",
            "x": "q-LsvU772uV_2sPJhfAIq-3vnKNVefNoIlvyvg1hrnE",
            "d": "Jcv8gklhMjC0b-lsk5onBbppWAx5ncNtbM63Jr9xBQE",
        }
    )

    protected = {
        "alg": "ECDH-1PU+A128KW",
        "enc": "A256CBC-HS512",
        "apu": "QWxpY2U",
        "apv": "Qm9iIGFuZCBDaGFybGll",
    }

    unprotected = {"jku": "https://provider.test/jwks"}

    recipients = [
        {
            "header": {
                "kid": "did:example:123#_Qq0UL2Fq651Q0Fjd6TvnYE-faHiOpRlPVQcY_-tA4A"
            }
        },
        {
            "header": {
                "kid": "did:example:123#_TKzHv2jFIyvdTGF1Dsgwngfdg3SH6TpDv0Ta1aOEkw"
            }
        },
    ]

    jwe_aad = b"Authenticate me too."

    header_obj = {
        "protected": protected,
        "unprotected": unprotected,
        "recipients": recipients,
        "aad": jwe_aad,
    }

    payload = b"Three is a magic number."

    data = jwe.serialize_json(
        header_obj, payload, [bob_key, charlie_key], sender_key=alice_key
    )

    rv_at_bob = jwe.deserialize_json(data, (bob_kid, bob_key), sender_key=alice_key)

    assert rv_at_bob["header"]["protected"].keys() == protected.keys() | {"epk"}
    assert {
        k: rv_at_bob["header"]["protected"][k]
        for k in rv_at_bob["header"]["protected"].keys() - {"epk"}
    } == protected
    assert rv_at_bob["header"]["unprotected"] == unprotected
    assert rv_at_bob["header"]["recipients"] == recipients
    assert rv_at_bob["header"]["aad"] == jwe_aad
    assert rv_at_bob["payload"] == payload

    rv_at_charlie = jwe.deserialize_json(
        data, (charlie_kid, charlie_key), sender_key=alice_key
    )

    assert rv_at_charlie["header"]["protected"].keys() == protected.keys() | {"epk"}
    assert {
        k: rv_at_charlie["header"]["protected"][k]
        for k in rv_at_charlie["header"]["protected"].keys() - {"epk"}
    } == protected
    assert rv_at_charlie["header"]["unprotected"] == unprotected
    assert rv_at_charlie["header"]["recipients"] == recipients
    assert rv_at_charlie["header"]["aad"] == jwe_aad
    assert rv_at_charlie["payload"] == payload


def test_ecdh_1pu_jwe_with_json_serialization_for_single_recipient():
    jwe = JsonWebEncryption()

    alice_key = OKPKey.import_key(
        {
            "kty": "OKP",
            "crv": "X25519",
            "x": "Knbm_BcdQr7WIoz-uqit9M0wbcfEr6y-9UfIZ8QnBD4",
            "d": "i9KuFhSzEBsiv3PKVL5115OCdsqQai5nj_Flzfkw5jU",
        }
    )
    bob_key = OKPKey.import_key(
        {
            "kty": "OKP",
            "crv": "X25519",
            "x": "BT7aR0ItXfeDAldeeOlXL_wXqp-j5FltT0vRSG16kRw",
            "d": "1gDirl_r_Y3-qUa3WXHgEXrrEHngWThU3c9zj9A2uBg",
        }
    )

    protected = {
        "alg": "ECDH-1PU+A128KW",
        "enc": "A256CBC-HS512",
        "apu": "QWxpY2U",
        "apv": "Qm9i",
    }

    unprotected = {"jku": "https://provider.test/jwks"}

    recipients = [{"header": {"kid": "bob-key-2"}}]

    jwe_aad = b"Authenticate me too."

    header_obj = {
        "protected": protected,
        "unprotected": unprotected,
        "recipients": recipients,
        "aad": jwe_aad,
    }

    payload = b"Three is a magic number."

    data = jwe.serialize_json(header_obj, payload, bob_key, sender_key=alice_key)

    rv = jwe.deserialize_json(data, bob_key, sender_key=alice_key)

    assert rv["header"]["protected"].keys() == protected.keys() | {"epk"}
    assert {
        k: rv["header"]["protected"][k]
        for k in rv["header"]["protected"].keys() - {"epk"}
    } == protected
    assert rv["header"]["unprotected"] == unprotected
    assert rv["header"]["recipients"] == recipients
    assert rv["header"]["aad"] == jwe_aad
    assert rv["payload"] == payload


def test_ecdh_1pu_encryption_fails_json_serialization_multiple_recipients_in_direct_key_agreement_mode():
    jwe = JsonWebEncryption()
    alice_key = OKPKey.generate_key("X25519", is_private=True)
    bob_key = OKPKey.generate_key("X25519", is_private=True)
    charlie_key = OKPKey.generate_key("X25519", is_private=True)

    protected = {"alg": "ECDH-1PU", "enc": "A128GCM"}
    header_obj = {"protected": protected}
    with pytest.raises(InvalidAlgorithmForMultipleRecipientsMode):
        jwe.serialize_json(
            header_obj,
            b"hello",
            [bob_key, charlie_key],
            sender_key=alice_key,
        )


def test_ecdh_1pu_encryption_fails_if_not_aes_cbc_hmac_sha2_enc_is_used_with_kw():
    jwe = JsonWebEncryption()
    alice_key = {
        "kty": "EC",
        "crv": "P-256",
        "x": "WKn-ZIGevcwGIyyrzFoZNBdaq9_TsqzGl96oc0CWuis",
        "y": "y77t-RvAHRKTsSGdIYUfweuOvwrvDD-Q3Hv5J0fSKbE",
        "d": "Hndv7ZZjs_ke8o9zXYo3iq-Yr8SewI5vrqd0pAvEPqg",
    }
    bob_key = {
        "kty": "EC",
        "crv": "P-256",
        "x": "weNJy2HscCSM6AEDTDg04biOvhFhyyWvOHQfeF_PxMQ",
        "y": "e8lnCO-AlStT-NJVX-crhB7QRYhiix03illJOVAOyck",
    }

    for alg in [
        "ECDH-1PU+A128KW",
        "ECDH-1PU+A192KW",
        "ECDH-1PU+A256KW",
    ]:
        for enc in [
            "A128GCM",
            "A192GCM",
            "A256GCM",
        ]:
            protected = {"alg": alg, "enc": enc}
            with pytest.raises(
                InvalidEncryptionAlgorithmForECDH1PUWithKeyWrappingError
            ):
                jwe.serialize_compact(
                    protected,
                    b"hello",
                    bob_key,
                    sender_key=alice_key,
                )


def test_ecdh_1pu_encryption_with_public_sender_key_fails():
    jwe = JsonWebEncryption()
    protected = {"alg": "ECDH-1PU", "enc": "A256GCM"}

    alice_key = {
        "kty": "EC",
        "crv": "P-256",
        "x": "WKn-ZIGevcwGIyyrzFoZNBdaq9_TsqzGl96oc0CWuis",
        "y": "y77t-RvAHRKTsSGdIYUfweuOvwrvDD-Q3Hv5J0fSKbE",
    }
    bob_key = {
        "kty": "EC",
        "crv": "P-256",
        "x": "weNJy2HscCSM6AEDTDg04biOvhFhyyWvOHQfeF_PxMQ",
        "y": "e8lnCO-AlStT-NJVX-crhB7QRYhiix03illJOVAOyck",
        "d": "VEmDZpDXXK8p8N0Cndsxs924q6nS1RXFASRl6BfUqdw",
    }
    with pytest.raises(ValueError):
        jwe.serialize_compact(
            protected,
            b"hello",
            bob_key,
            sender_key=alice_key,
        )


def test_ecdh_1pu_decryption_with_public_recipient_key_fails():
    jwe = JsonWebEncryption()
    protected = {"alg": "ECDH-1PU", "enc": "A256GCM"}

    alice_key = {
        "kty": "EC",
        "crv": "P-256",
        "x": "WKn-ZIGevcwGIyyrzFoZNBdaq9_TsqzGl96oc0CWuis",
        "y": "y77t-RvAHRKTsSGdIYUfweuOvwrvDD-Q3Hv5J0fSKbE",
        "d": "Hndv7ZZjs_ke8o9zXYo3iq-Yr8SewI5vrqd0pAvEPqg",
    }
    bob_key = {
        "kty": "EC",
        "crv": "P-256",
        "x": "weNJy2HscCSM6AEDTDg04biOvhFhyyWvOHQfeF_PxMQ",
        "y": "e8lnCO-AlStT-NJVX-crhB7QRYhiix03illJOVAOyck",
    }
    data = jwe.serialize_compact(protected, b"hello", bob_key, sender_key=alice_key)
    with pytest.raises(ValueError):
        jwe.deserialize_compact(data, bob_key, sender_key=alice_key)


def test_ecdh_1pu_encryption_fails_if_key_types_are_different():
    jwe = JsonWebEncryption()
    protected = {"alg": "ECDH-1PU", "enc": "A256GCM"}

    alice_key = ECKey.generate_key("P-256", is_private=True)
    bob_key = OKPKey.generate_key("X25519", is_private=False)
    with pytest.raises(TypeError):
        jwe.serialize_compact(
            protected,
            b"hello",
            bob_key,
            sender_key=alice_key,
        )

    alice_key = OKPKey.generate_key("X25519", is_private=True)
    bob_key = ECKey.generate_key("P-256", is_private=False)
    with pytest.raises(TypeError):
        jwe.serialize_compact(
            protected,
            b"hello",
            bob_key,
            sender_key=alice_key,
        )


def test_ecdh_1pu_encryption_fails_if_keys_curves_are_different():
    jwe = JsonWebEncryption()
    protected = {"alg": "ECDH-1PU", "enc": "A256GCM"}

    alice_key = ECKey.generate_key("P-256", is_private=True)
    bob_key = ECKey.generate_key("secp256k1", is_private=False)
    with pytest.raises(ValueError):
        jwe.serialize_compact(
            protected,
            b"hello",
            bob_key,
            sender_key=alice_key,
        )

    alice_key = ECKey.generate_key("P-384", is_private=True)
    bob_key = ECKey.generate_key("P-521", is_private=False)
    with pytest.raises(ValueError):
        jwe.serialize_compact(
            protected,
            b"hello",
            bob_key,
            sender_key=alice_key,
        )

    alice_key = OKPKey.generate_key("X25519", is_private=True)
    bob_key = OKPKey.generate_key("X448", is_private=False)
    with pytest.raises(TypeError):
        jwe.serialize_compact(
            protected,
            b"hello",
            bob_key,
            sender_key=alice_key,
        )


def test_ecdh_1pu_encryption_fails_if_key_points_are_not_actually_on_same_curve():
    jwe = JsonWebEncryption()
    protected = {"alg": "ECDH-1PU", "enc": "A256GCM"}

    alice_key = {
        "kty": "EC",
        "crv": "P-256",
        "x": "aDHtGkIYyhR5geqfMaFL0T9cG4JEMI8nyMFJA7gRUDs",
        "y": "AjGN5_f-aCt4vYg74my6n1ALIq746nlc_httIgcBSYY",
        "d": "Sim3EIzXsWaWu9QW8yKVHwxBM5CTlnrVU_Eq-y_KRQA",
    }  # the point is indeed on P-256 curve
    bob_key = {
        "kty": "EC",
        "crv": "P-256",
        "x": "5ZFnZbs_BtLBIZxwt5hS7SBDtI2a-dJ871dJ8ZnxZ6c",
        "y": "K0srqSkbo1Yeckr0YoQA8r_rOz0ZUStiv3mc1qn46pg",
    }  # the point is not on P-256 curve but is actually on secp256k1 curve

    with pytest.raises(ValueError):
        jwe.serialize_compact(
            protected,
            b"hello",
            bob_key,
            sender_key=alice_key,
        )

    alice_key = {
        "kty": "EC",
        "crv": "P-521",
        "x": "1JDMOjnMgASo01PVHRcyCDtE6CLgKuwXLXLbdLGxpdubLuHYBa0KAepyimnxCWsX",
        "y": "w7BSC8Xb3XgMMfE7IFCJpoOmx1Sf3T3_3OZ4CrF6_iCFAw4VOdFYR42OnbKMFG--",
        "d": "lCkpFBaVwHzfHtkJEV3PzxefObOPnMgUjNZSLryqC5AkERgXT3-DZLEi6eBzq5gk",
    }  # the point is not on P-521 curve but is actually on P-384 curve
    bob_key = {
        "kty": "EC",
        "crv": "P-521",
        "x": "Cd6rinJdgS4WJj6iaNyXiVhpMbhZLmPykmrnFhIad04B3ulf5pURb5v9mx21c_Cv8Q1RBOptwleLg5Qjq2J1qa4",
        "y": "hXo9p1EjW6W4opAQdmfNgyxztkNxYwn9L4FVTLX51KNEsW0aqueLm96adRmf0HoGIbNhIdcIlXOKlRUHqgunDkM",
    }  # the point is indeed on P-521 curve

    with pytest.raises(ValueError):
        jwe.serialize_compact(
            protected,
            b"hello",
            bob_key,
            sender_key=alice_key,
        )

    alice_key = OKPKey.import_key(
        {
            "kty": "OKP",
            "crv": "X25519",
            "x": "TAB1oIsjPob3guKwTEeQsAsupSRPdXdxHhnV8JrVJTA",
            "d": "kO2LzPr4vLg_Hn-7_MDq66hJZgvTIkzDG4p6nCsgNHk",
        }
    )  # the point is indeed on X25519 curve
    bob_key = OKPKey.import_key(
        {
            "kty": "OKP",
            "crv": "X25519",
            "x": "lVHcPx4R9bExaoxXZY9tAq7SNW9pJKCoVQxURLtkAs3Dg5ZRxcjhf0JUyg2lod5OGDptJ7wowwY",
        }
    )  # the point is not on X25519 curve but is actually on X448 curve

    with pytest.raises(ValueError):
        jwe.serialize_compact(
            protected,
            b"hello",
            bob_key,
            sender_key=alice_key,
        )

    alice_key = OKPKey.import_key(
        {
            "kty": "OKP",
            "crv": "X448",
            "x": "TAB1oIsjPob3guKwTEeQsAsupSRPdXdxHhnV8JrVJTA",
            "d": "kO2LzPr4vLg_Hn-7_MDq66hJZgvTIkzDG4p6nCsgNHk",
        }
    )  # the point is not on X448 curve but is actually on X25519 curve
    bob_key = OKPKey.import_key(
        {
            "kty": "OKP",
            "crv": "X448",
            "x": "lVHcPx4R9bExaoxXZY9tAq7SNW9pJKCoVQxURLtkAs3Dg5ZRxcjhf0JUyg2lod5OGDptJ7wowwY",
        }
    )  # the point is indeed on X448 curve

    with pytest.raises(ValueError):
        jwe.serialize_compact(
            protected,
            b"hello",
            bob_key,
            sender_key=alice_key,
        )


def test_ecdh_1pu_encryption_fails_if_keys_curve_is_inappropriate():
    jwe = JsonWebEncryption()
    protected = {"alg": "ECDH-1PU", "enc": "A256GCM"}

    alice_key = OKPKey.generate_key(
        "Ed25519", is_private=True
    )  # use Ed25519 instead of X25519
    bob_key = OKPKey.generate_key(
        "Ed25519", is_private=False
    )  # use Ed25519 instead of X25519
    with pytest.raises(ValueError):
        jwe.serialize_compact(
            protected,
            b"hello",
            bob_key,
            sender_key=alice_key,
        )


def test_ecdh_1pu_encryption_for_multiple_recipients_fails_if_key_types_are_different():
    jwe = JsonWebEncryption()
    protected = {"alg": "ECDH-1PU+A128KW", "enc": "A128CBC-HS256"}
    header_obj = {"protected": protected}

    alice_key = ECKey.generate_key("P-256", is_private=True)
    bob_key = ECKey.generate_key("P-256", is_private=False)
    charlie_key = OKPKey.generate_key("X25519", is_private=False)

    with pytest.raises(TypeError):
        jwe.serialize_json(
            header_obj,
            b"hello",
            [bob_key, charlie_key],
            sender_key=alice_key,
        )


def test_ecdh_1pu_encryption_for_multiple_recipients_fails_if_keys_curves_are_different():
    jwe = JsonWebEncryption()
    protected = {"alg": "ECDH-1PU+A128KW", "enc": "A128CBC-HS256"}
    header_obj = {"protected": protected}

    alice_key = OKPKey.generate_key("X25519", is_private=True)
    bob_key = OKPKey.generate_key("X448", is_private=False)
    charlie_key = OKPKey.generate_key("X25519", is_private=False)

    with pytest.raises(TypeError):
        jwe.serialize_json(
            header_obj,
            b"hello",
            [bob_key, charlie_key],
            sender_key=alice_key,
        )


def test_ecdh_1pu_encryption_for_multiple_recipients_fails_if_key_points_are_not_actually_on_same_curve():
    jwe = JsonWebEncryption()
    protected = {"alg": "ECDH-1PU+A128KW", "enc": "A128CBC-HS256"}
    header_obj = {"protected": protected}

    alice_key = {
        "kty": "EC",
        "crv": "P-256",
        "x": "aDHtGkIYyhR5geqfMaFL0T9cG4JEMI8nyMFJA7gRUDs",
        "y": "AjGN5_f-aCt4vYg74my6n1ALIq746nlc_httIgcBSYY",
        "d": "Sim3EIzXsWaWu9QW8yKVHwxBM5CTlnrVU_Eq-y_KRQA",
    }  # the point is indeed on P-256 curve
    bob_key = {
        "kty": "EC",
        "crv": "P-256",
        "x": "HgF88mm6yw4gjG7yG6Sqz66pHnpZcyx7c842BQghYuc",
        "y": "KZ1ywvTOYnpNb4Gepa5eSgfEOb5gj5hCaCFIrTFuI2o",
    }  # the point is indeed on P-256 curve
    charlie_key = {
        "kty": "EC",
        "crv": "P-256",
        "x": "5ZFnZbs_BtLBIZxwt5hS7SBDtI2a-dJ871dJ8ZnxZ6c",
        "y": "K0srqSkbo1Yeckr0YoQA8r_rOz0ZUStiv3mc1qn46pg",
    }  # the point is not on P-256 curve but is actually on secp256k1 curve

    with pytest.raises(ValueError):
        jwe.serialize_json(
            header_obj,
            b"hello",
            [bob_key, charlie_key],
            sender_key=alice_key,
        )


def test_ecdh_1pu_encryption_for_multiple_recipients_fails_if_keys_curve_is_inappropriate():
    jwe = JsonWebEncryption()
    protected = {"alg": "ECDH-1PU+A128KW", "enc": "A128CBC-HS256"}
    header_obj = {"protected": protected}

    alice_key = OKPKey.generate_key(
        "Ed25519", is_private=True
    )  # use Ed25519 instead of X25519
    bob_key = OKPKey.generate_key(
        "Ed25519", is_private=False
    )  # use Ed25519 instead of X25519
    charlie_key = OKPKey.generate_key(
        "Ed25519", is_private=False
    )  # use Ed25519 instead of X25519

    with pytest.raises(ValueError):
        jwe.serialize_json(
            header_obj,
            b"hello",
            [bob_key, charlie_key],
            sender_key=alice_key,
        )


def test_ecdh_1pu_decryption_fails_if_key_matches_to_no_recipient():
    jwe = JsonWebEncryption()

    alice_key = OKPKey.import_key(
        {
            "kty": "OKP",
            "crv": "X25519",
            "x": "Knbm_BcdQr7WIoz-uqit9M0wbcfEr6y-9UfIZ8QnBD4",
            "d": "i9KuFhSzEBsiv3PKVL5115OCdsqQai5nj_Flzfkw5jU",
        }
    )
    bob_key = OKPKey.import_key(
        {
            "kty": "OKP",
            "crv": "X25519",
            "x": "BT7aR0ItXfeDAldeeOlXL_wXqp-j5FltT0vRSG16kRw",
            "d": "1gDirl_r_Y3-qUa3WXHgEXrrEHngWThU3c9zj9A2uBg",
        }
    )
    charlie_key = OKPKey.import_key(
        {
            "kty": "OKP",
            "crv": "X25519",
            "x": "q-LsvU772uV_2sPJhfAIq-3vnKNVefNoIlvyvg1hrnE",
            "d": "Jcv8gklhMjC0b-lsk5onBbppWAx5ncNtbM63Jr9xBQE",
        }
    )

    protected = {
        "alg": "ECDH-1PU+A128KW",
        "enc": "A256CBC-HS512",
        "apu": "QWxpY2U",
        "apv": "Qm9i",
    }

    unprotected = {"jku": "https://provider.test/jwks"}

    recipients = [{"header": {"kid": "bob-key-2"}}]

    jwe_aad = b"Authenticate me too."

    header_obj = {
        "protected": protected,
        "unprotected": unprotected,
        "recipients": recipients,
        "aad": jwe_aad,
    }

    payload = b"Three is a magic number."

    data = jwe.serialize_json(header_obj, payload, bob_key, sender_key=alice_key)

    with pytest.raises(InvalidUnwrap):
        jwe.deserialize_json(data, charlie_key, sender_key=alice_key)



from collections import OrderedDict
from copy import deepcopy

from authlib.common.encoding import json_b64encode
from authlib.common.encoding import to_bytes
from authlib.common.encoding import to_unicode
from authlib.common.encoding import urlsafe_b64encode
from authlib.jose.errors import DecodeError
from authlib.jose.errors import InvalidAlgorithmForMultipleRecipientsMode
from authlib.jose.errors import InvalidHeaderParameterNameError
from authlib.jose.errors import KeyMismatchError
from authlib.jose.errors import MissingAlgorithmError
from authlib.jose.errors import MissingEncryptionAlgorithmError
from authlib.jose.errors import UnsupportedAlgorithmError
from authlib.jose.errors import UnsupportedCompressionAlgorithmError
from authlib.jose.errors import UnsupportedEncryptionAlgorithmError
from authlib.jose.rfc7516.models import JWEAlgorithmWithTagAwareKeyAgreement
from authlib.jose.rfc7516.models import JWEHeader
from authlib.jose.rfc7516.models import JWESharedHeader
from authlib.jose.util import ensure_dict
from authlib.jose.util import extract_header
from authlib.jose.util import extract_segment


class JsonWebEncryption:
    #: Registered Header Parameter Names defined by Section 4.1
    REGISTERED_HEADER_PARAMETER_NAMES = frozenset(
        [
            "alg",
            "enc",
            "zip",
            "jku",
            "jwk",
            "kid",
            "x5u",
            "x5c",
            "x5t",
            "x5t#S256",
            "typ",
            "cty",
            "crit",
        ]
    )

    ALG_REGISTRY = {}
    ENC_REGISTRY = {}
    ZIP_REGISTRY = {}

    def __init__(self, algorithms=None, private_headers=None):
        self._algorithms = algorithms
        self._private_headers = private_headers

    @classmethod
    def register_algorithm(cls, algorithm):
        """Register an algorithm for ``alg`` or ``enc`` or ``zip`` of JWE."""
        if not algorithm or algorithm.algorithm_type != "JWE":
            raise ValueError(f"Invalid algorithm for JWE, {algorithm!r}")

        if algorithm.algorithm_location == "alg":
            cls.ALG_REGISTRY[algorithm.name] = algorithm
        elif algorithm.algorithm_location == "enc":
            cls.ENC_REGISTRY[algorithm.name] = algorithm
        elif algorithm.algorithm_location == "zip":
            cls.ZIP_REGISTRY[algorithm.name] = algorithm

    def serialize_compact(self, protected, payload, key, sender_key=None):
        """Generate a JWE Compact Serialization.

        The JWE Compact Serialization represents encrypted content as a compact,
        URL-safe string. This string is::

            BASE64URL(UTF8(JWE Protected Header)) || '.' ||
            BASE64URL(JWE Encrypted Key) || '.' ||
            BASE64URL(JWE Initialization Vector) || '.' ||
            BASE64URL(JWE Ciphertext) || '.' ||
            BASE64URL(JWE Authentication Tag)

        Only one recipient is supported by the JWE Compact Serialization and
        it provides no syntax to represent JWE Shared Unprotected Header, JWE
        Per-Recipient Unprotected Header, or JWE AAD values.

        :param protected: A dict of protected header
        :param payload: Payload (bytes or a value convertible to bytes)
        :param key: Public key used to encrypt payload
        :param sender_key: Sender's private key in case
            JWEAlgorithmWithTagAwareKeyAgreement is used
        :return: JWE compact serialization as bytes
        """
        # step 1: Prepare algorithms & key
        alg = self.get_header_alg(protected)
        enc = self.get_header_enc(protected)
        zip_alg = self.get_header_zip(protected)

        self._validate_sender_key(sender_key, alg)
        self._validate_private_headers(protected, alg)

        key = prepare_key(alg, protected, key)
        if sender_key is not None:
            sender_key = alg.prepare_key(sender_key)

        # self._post_validate_header(protected, algorithm)

        # step 2: Generate a random Content Encryption Key (CEK)
        # use enc_alg.generate_cek() in scope of upcoming .wrap
        # or .generate_keys_and_prepare_headers call

        # step 3: Encrypt the CEK with the recipient's public key
        if (
            isinstance(alg, JWEAlgorithmWithTagAwareKeyAgreement)
            and alg.key_size is not None
        ):
            # For a JWE algorithm with tag-aware key agreement in case key agreement
            # with key wrapping mode is used:
            # Defer key agreement with key wrapping until
            # authentication tag is computed
            prep = alg.generate_keys_and_prepare_headers(enc, key, sender_key)
            epk = prep["epk"]
            cek = prep["cek"]
            protected.update(prep["header"])
        else:
            # In any other case:
            # Keep the normal steps order defined by RFC 7516
            if isinstance(alg, JWEAlgorithmWithTagAwareKeyAgreement):
                wrapped = alg.wrap(enc, protected, key, sender_key)
            else:
                wrapped = alg.wrap(enc, protected, key)
            cek = wrapped["cek"]
            ek = wrapped["ek"]
            if "header" in wrapped:
                protected.update(wrapped["header"])

        # step 4: Generate a random JWE Initialization Vector
        iv = enc.generate_iv()

        # step 5: Let the Additional Authenticated Data encryption parameter
        # be ASCII(BASE64URL(UTF8(JWE Protected Header)))
        protected_segment = json_b64encode(protected)
        aad = to_bytes(protected_segment, "ascii")

        # step 6: compress message if required
        if zip_alg:
            msg = zip_alg.compress(to_bytes(payload))
        else:
            msg = to_bytes(payload)

        # step 7: perform encryption
        ciphertext, tag = enc.encrypt(msg, aad, iv, cek)

        if (
            isinstance(alg, JWEAlgorithmWithTagAwareKeyAgreement)
            and alg.key_size is not None
        ):
            # For a JWE algorithm with tag-aware key agreement in case key agreement
            # with key wrapping mode is used:
            # Perform key agreement with key wrapping deferred at step 3
            wrapped = alg.agree_upon_key_and_wrap_cek(
                enc, protected, key, sender_key, epk, cek, tag
            )
            ek = wrapped["ek"]

        # step 8: build resulting message
        return b".".join(
            [
                protected_segment,
                urlsafe_b64encode(ek),
                urlsafe_b64encode(iv),
                urlsafe_b64encode(ciphertext),
                urlsafe_b64encode(tag),
            ]
        )

    def serialize_json(self, header_obj, payload, keys, sender_key=None):  # noqa: C901
        """Generate a JWE JSON Serialization (in fully general syntax).

        The JWE JSON Serialization represents encrypted content as a JSON
        object.  This representation is neither optimized for compactness nor
        URL safe.

        The following members are defined for use in top-level JSON objects
        used for the fully general JWE JSON Serialization syntax:

        protected
            The "protected" member MUST be present and contain the value
            BASE64URL(UTF8(JWE Protected Header)) when the JWE Protected
            Header value is non-empty; otherwise, it MUST be absent.  These
            Header Parameter values are integrity protected.

        unprotected
            The "unprotected" member MUST be present and contain the value JWE
            Shared Unprotected Header when the JWE Shared Unprotected Header
            value is non-empty; otherwise, it MUST be absent.  This value is
            represented as an unencoded JSON object, rather than as a string.
            These Header Parameter values are not integrity protected.

        iv
            The "iv" member MUST be present and contain the value
            BASE64URL(JWE Initialization Vector) when the JWE Initialization
            Vector value is non-empty; otherwise, it MUST be absent.

        aad
            The "aad" member MUST be present and contain the value
            BASE64URL(JWE AAD)) when the JWE AAD value is non-empty;
            otherwise, it MUST be absent.  A JWE AAD value can be included to
            supply a base64url-encoded value to be integrity protected but not
            encrypted.

        ciphertext
            The "ciphertext" member MUST be present and contain the value
            BASE64URL(JWE Ciphertext).

        tag
            The "tag" member MUST be present and contain the value
            BASE64URL(JWE Authentication Tag) when the JWE Authentication Tag
            value is non-empty; otherwise, it MUST be absent.

        recipients
            The "recipients" member value MUST be an array of JSON objects.
            Each object contains information specific to a single recipient.
            This member MUST be present with exactly one array element per
            recipient, even if some or all of the array element values are the
            empty JSON object "{}" (which can happen when all Header Parameter
            values are shared between all recipients and when no encrypted key
            is used, such as when doing Direct Encryption).

        The following members are defined for use in the JSON objects that
        are elements of the "recipients" array:

        header
            The "header" member MUST be present and contain the value JWE Per-
            Recipient Unprotected Header when the JWE Per-Recipient
            Unprotected Header value is non-empty; otherwise, it MUST be
            absent.  This value is represented as an unencoded JSON object,
            rather than as a string.  These Header Parameter values are not
            integrity protected.

        encrypted_key
            The "encrypted_key" member MUST be present and contain the value
            BASE64URL(JWE Encrypted Key) when the JWE Encrypted Key value is
            non-empty; otherwise, it MUST be absent.

        This implementation assumes that "alg" and "enc" header fields are
        contained in the protected or shared unprotected header.

        :param header_obj: A dict of headers (in addition optionally contains JWE AAD)
        :param payload: Payload (bytes or a value convertible to bytes)
        :param keys: Public keys (or a single public key) used to encrypt payload
        :param sender_key: Sender's private key in case
            JWEAlgorithmWithTagAwareKeyAgreement is used
        :return: JWE JSON serialization (in fully general syntax) as dict

        Example of `header_obj`::

            {
                "protected": {
                    "alg": "ECDH-1PU+A128KW",
                    "enc": "A256CBC-HS512",
                    "apu": "QWxpY2U",
                    "apv": "Qm9iIGFuZCBDaGFybGll",
                },
                "unprotected": {"jku": "https://alice.example.com/keys.jwks"},
                "recipients": [
                    {"header": {"kid": "bob-key-2"}},
                    {"header": {"kid": "2021-05-06"}},
                ],
                "aad": b"Authenticate me too.",
            }
        """
        if not isinstance(keys, list):  # single key
            keys = [keys]

        if not keys:
            raise ValueError("No keys have been provided")

        header_obj = deepcopy(header_obj)

        shared_header = JWESharedHeader.from_dict(header_obj)

        recipients = header_obj.get("recipients")
        if recipients is None:
            recipients = [{} for _ in keys]
        for i in range(len(recipients)):
            if recipients[i] is None:
                recipients[i] = {}
            if "header" not in recipients[i]:
                recipients[i]["header"] = {}

        jwe_aad = header_obj.get("aad")

        if len(keys) != len(recipients):
            raise ValueError(
                f"Count of recipient keys {len(keys)} does not equal to count of recipients {len(recipients)}"
            )

        # step 1: Prepare algorithms & key
        alg = self.get_header_alg(shared_header)
        enc = self.get_header_enc(shared_header)
        zip_alg = self.get_header_zip(shared_header)

        self._validate_sender_key(sender_key, alg)
        self._validate_private_headers(shared_header, alg)
        for recipient in recipients:
            self._validate_private_headers(recipient["header"], alg)

        for i in range(len(keys)):
            keys[i] = prepare_key(alg, recipients[i]["header"], keys[i])
        if sender_key is not None:
            sender_key = alg.prepare_key(sender_key)

        # self._post_validate_header(protected, algorithm)

        # step 2: Generate a random Content Encryption Key (CEK)
        # use enc_alg.generate_cek() in scope of upcoming .wrap
        # or .generate_keys_and_prepare_headers call

        # step 3: Encrypt the CEK with the recipient's public key
        preset = alg.generate_preset(enc, keys[0])
        if "cek" in preset:
            cek = preset["cek"]
        else:
            cek = None
        if len(keys) > 1 and cek is None:
            raise InvalidAlgorithmForMultipleRecipientsMode(alg.name)
        if "header" in preset:
            shared_header.update_protected(preset["header"])

        if (
            isinstance(alg, JWEAlgorithmWithTagAwareKeyAgreement)
            and alg.key_size is not None
        ):
            # For a JWE algorithm with tag-aware key agreement in case key agreement
            # with key wrapping mode is used:
            # Defer key agreement with key wrapping until authentication tag is computed
            epks = []
            for i in range(len(keys)):
                prep = alg.generate_keys_and_prepare_headers(
                    enc, keys[i], sender_key, preset
                )
                if cek is None:
                    cek = prep["cek"]
                epks.append(prep["epk"])
                recipients[i]["header"].update(prep["header"])
        else:
            # In any other case:
            # Keep the normal steps order defined by RFC 7516
            for i in range(len(keys)):
                if isinstance(alg, JWEAlgorithmWithTagAwareKeyAgreement):
                    wrapped = alg.wrap(enc, shared_header, keys[i], sender_key, preset)
                else:
                    wrapped = alg.wrap(enc, shared_header, keys[i], preset)
                if cek is None:
                    cek = wrapped["cek"]
                recipients[i]["encrypted_key"] = wrapped["ek"]
                if "header" in wrapped:
                    recipients[i]["header"].update(wrapped["header"])

        # step 4: Generate a random JWE Initialization Vector
        iv = enc.generate_iv()

        # step 5: Compute the Encoded Protected Header value
        # BASE64URL(UTF8(JWE Protected Header)). If the JWE Protected Header
        # is not present, let this value be the empty string.
        # Let the Additional Authenticated Data encryption parameter be
        # ASCII(Encoded Protected Header). However, if a JWE AAD value is
        # present, instead let the Additional Authenticated Data encryption
        # parameter be ASCII(Encoded Protected Header || '.' || BASE64URL(JWE AAD)).
        aad = (
            json_b64encode(shared_header.protected) if shared_header.protected else b""
        )
        if jwe_aad is not None:
            aad += b"." + urlsafe_b64encode(jwe_aad)
        aad = to_bytes(aad, "ascii")

        # step 6: compress message if required
        if zip_alg:
            msg = zip_alg.compress(to_bytes(payload))
        else:
            msg = to_bytes(payload)

        # step 7: perform encryption
        ciphertext, tag = enc.encrypt(msg, aad, iv, cek)

        if (
            isinstance(alg, JWEAlgorithmWithTagAwareKeyAgreement)
            and alg.key_size is not None
        ):
            # For a JWE algorithm with tag-aware key agreement in case key agreement
            # with key wrapping mode is used:
            # Perform key agreement with key wrapping deferred at step 3
            for i in range(len(keys)):
                wrapped = alg.agree_upon_key_and_wrap_cek(
                    enc, shared_header, keys[i], sender_key, epks[i], cek, tag
                )
                recipients[i]["encrypted_key"] = wrapped["ek"]

        # step 8: build resulting message
        obj = OrderedDict()

        if shared_header.protected:
            obj["protected"] = to_unicode(json_b64encode(shared_header.protected))

        if shared_header.unprotected:
            obj["unprotected"] = shared_header.unprotected

        for recipient in recipients:
            if not recipient["header"]:
                del recipient["header"]
            recipient["encrypted_key"] = to_unicode(
                urlsafe_b64encode(recipient["encrypted_key"])
            )
            for member in set(recipient.keys()):
                if member not in {"header", "encrypted_key"}:
                    del recipient[member]
        obj["recipients"] = recipients

        if jwe_aad is not None:
            obj["aad"] = to_unicode(urlsafe_b64encode(jwe_aad))

        obj["iv"] = to_unicode(urlsafe_b64encode(iv))

        obj["ciphertext"] = to_unicode(urlsafe_b64encode(ciphertext))

        obj["tag"] = to_unicode(urlsafe_b64encode(tag))

        return obj

    def serialize(self, header, payload, key, sender_key=None):
        """Generate a JWE Serialization.

        It will automatically generate a compact or JSON serialization depending
        on `header` argument. If `header` is a dict with "protected",
        "unprotected" and/or "recipients" keys, it will call `serialize_json`,
        otherwise it will call `serialize_compact`.

        :param header: A dict of header(s)
        :param payload: Payload (bytes or a value convertible to bytes)
        :param key: Public key(s) used to encrypt payload
        :param sender_key: Sender's private key in case
            JWEAlgorithmWithTagAwareKeyAgreement is used
        :return: JWE compact serialization as bytes or
            JWE JSON serialization as dict
        """
        if "protected" in header or "unprotected" in header or "recipients" in header:
            return self.serialize_json(header, payload, key, sender_key)

        return self.serialize_compact(header, payload, key, sender_key)

    def deserialize_compact(self, s, key, decode=None, sender_key=None):
        """Extract JWE Compact Serialization.

        :param s: JWE Compact Serialization as bytes
        :param key: Private key used to decrypt payload
            (optionally can be a tuple of kid and essentially key)
        :param decode: Function to decode payload data
        :param sender_key: Sender's public key in case
            JWEAlgorithmWithTagAwareKeyAgreement is used
        :return: dict with `header` and `payload` keys where `header` value is
            a dict containing protected header fields
        """
        try:
            s = to_bytes(s)
            protected_s, ek_s, iv_s, ciphertext_s, tag_s = s.rsplit(b".")
        except ValueError as exc:
            raise DecodeError("Not enough segments") from exc

        protected = extract_header(protected_s, DecodeError)
        ek = extract_segment(ek_s, DecodeError, "encryption key")
        iv = extract_segment(iv_s, DecodeError, "initialization vector")
        ciphertext = extract_segment(ciphertext_s, DecodeError, "ciphertext")
        tag = extract_segment(tag_s, DecodeError, "authentication tag")

        alg = self.get_header_alg(protected)
        enc = self.get_header_enc(protected)
        zip_alg = self.get_header_zip(protected)

        self._validate_sender_key(sender_key, alg)
        self._validate_private_headers(protected, alg)

        if isinstance(key, tuple) and len(key) == 2:
            # Ignore separately provided kid, extract essentially key only
            key = key[1]

        key = prepare_key(alg, protected, key)

        if sender_key is not None:
            sender_key = alg.prepare_key(sender_key)

        if isinstance(alg, JWEAlgorithmWithTagAwareKeyAgreement):
            # For a JWE algorithm with tag-aware key agreement:
            if alg.key_size is not None:
                # In case key agreement with key wrapping mode is used:
                # Provide authentication tag to .unwrap method
                cek = alg.unwrap(enc, ek, protected, key, sender_key, tag)
            else:
                # Otherwise, don't provide authentication tag to .unwrap method
                cek = alg.unwrap(enc, ek, protected, key, sender_key)
        else:
            # For any other JWE algorithm:
            # Don't provide authentication tag to .unwrap method
            cek = alg.unwrap(enc, ek, protected, key)

        aad = to_bytes(protected_s, "ascii")
        msg = enc.decrypt(ciphertext, aad, iv, tag, cek)

        if zip_alg:
            payload = zip_alg.decompress(to_bytes(msg))
        else:
            payload = msg

        if decode:
            payload = decode(payload)
        return {"header": protected, "payload": payload}

    def deserialize_json(self, obj, key, decode=None, sender_key=None):  # noqa: C901
        """Extract JWE JSON Serialization.

        :param obj: JWE JSON Serialization as dict or str
        :param key: Private key used to decrypt payload
            (optionally can be a tuple of kid and essentially key)
        :param decode: Function to decode payload data
        :param sender_key: Sender's public key in case
            JWEAlgorithmWithTagAwareKeyAgreement is used
        :return: dict with `header` and `payload` keys where `header` value is
            a dict containing `protected`, `unprotected`, `recipients` and/or
            `aad` keys
        """
        obj = ensure_dict(obj, "JWE")
        obj = deepcopy(obj)

        if "protected" in obj:
            protected = extract_header(to_bytes(obj["protected"]), DecodeError)
        else:
            protected = None

        unprotected = obj.get("unprotected")

        recipients = obj["recipients"]
        for recipient in recipients:
            if "header" not in recipient:
                recipient["header"] = {}
            recipient["encrypted_key"] = extract_segment(
                to_bytes(recipient["encrypted_key"]), DecodeError, "encrypted key"
            )

        if "aad" in obj:
            jwe_aad = extract_segment(to_bytes(obj["aad"]), DecodeError, "JWE AAD")
        else:
            jwe_aad = None

        iv = extract_segment(to_bytes(obj["iv"]), DecodeError, "initialization vector")

        ciphertext = extract_segment(
            to_bytes(obj["ciphertext"]), DecodeError, "ciphertext"
        )

        tag = extract_segment(to_bytes(obj["tag"]), DecodeError, "authentication tag")

        shared_header = JWESharedHeader(protected, unprotected)

        alg = self.get_header_alg(shared_header)
        enc = self.get_header_enc(shared_header)
        zip_alg = self.get_header_zip(shared_header)

        self._validate_sender_key(sender_key, alg)
        self._validate_private_headers(shared_header, alg)
        for recipient in recipients:
            self._validate_private_headers(recipient["header"], alg)

        kid = None
        if isinstance(key, tuple) and len(key) == 2:
            # Extract separately provided kid and essentially key
            kid = key[0]
            key = key[1]

        key = alg.prepare_key(key)

        if kid is None:
            # If kid has not been provided separately, try to get it from key itself
            kid = key.kid

        if sender_key is not None:
            sender_key = alg.prepare_key(sender_key)

        def _unwrap_with_sender_key_and_tag(ek, header):
            return alg.unwrap(enc, ek, header, key, sender_key, tag)

        def _unwrap_with_sender_key_and_without_tag(ek, header):
            return alg.unwrap(enc, ek, header, key, sender_key)

        def _unwrap_without_sender_key_and_tag(ek, header):
            return alg.unwrap(enc, ek, header, key)

        def _unwrap_for_matching_recipient(unwrap_func):
            if kid is not None:
                for recipient in recipients:
                    if recipient["header"].get("kid") == kid:
                        header = JWEHeader(protected, unprotected, recipient["header"])
                        return unwrap_func(recipient["encrypted_key"], header)

            # Since no explicit match has been found, iterate over all the recipients
            error = None
            for recipient in recipients:
                header = JWEHeader(protected, unprotected, recipient["header"])
                try:
                    return unwrap_func(recipient["encrypted_key"], header)
                except Exception as e:
                    error = e
            else:
                if error is None:
                    raise KeyMismatchError()
                else:
                    raise error

        if isinstance(alg, JWEAlgorithmWithTagAwareKeyAgreement):
            # For a JWE algorithm with tag-aware key agreement:
            if alg.key_size is not None:
                # In case key agreement with key wrapping mode is used:
                # Provide authentication tag to .unwrap method
                cek = _unwrap_for_matching_recipient(_unwrap_with_sender_key_and_tag)
            else:
                # Otherwise, don't provide authentication tag to .unwrap method
                cek = _unwrap_for_matching_recipient(
                    _unwrap_with_sender_key_and_without_tag
                )
        else:
            # For any other JWE algorithm:
            # Don't provide authentication tag to .unwrap method
            cek = _unwrap_for_matching_recipient(_unwrap_without_sender_key_and_tag)

        aad = to_bytes(obj.get("protected", ""))
        if "aad" in obj:
            aad += b"." + to_bytes(obj["aad"])
        aad = to_bytes(aad, "ascii")

        msg = enc.decrypt(ciphertext, aad, iv, tag, cek)

        if zip_alg:
            payload = zip_alg.decompress(to_bytes(msg))
        else:
            payload = msg

        if decode:
            payload = decode(payload)

        for recipient in recipients:
            if not recipient["header"]:
                del recipient["header"]
            for member in set(recipient.keys()):
                if member != "header":
                    del recipient[member]

        header = {}
        if protected:
            header["protected"] = protected
        if unprotected:
            header["unprotected"] = unprotected
        header["recipients"] = recipients
        if jwe_aad is not None:
            header["aad"] = jwe_aad

        return {"header": header, "payload": payload}

    def deserialize(self, obj, key, decode=None, sender_key=None):
        """Extract a JWE Serialization.

        It supports both compact and JSON serialization.

        :param obj: JWE compact serialization as bytes or
            JWE JSON serialization as dict or str
        :param key: Private key used to decrypt payload
            (optionally can be a tuple of kid and essentially key)
        :param decode: Function to decode payload data
        :param sender_key: Sender's public key in case
            JWEAlgorithmWithTagAwareKeyAgreement is used
        :return: dict with `header` and `payload` keys
        """
        if isinstance(obj, dict):
            return self.deserialize_json(obj, key, decode, sender_key)

        obj = to_bytes(obj)
        if obj.startswith(b"{") and obj.endswith(b"}"):
            return self.deserialize_json(obj, key, decode, sender_key)

        return self.deserialize_compact(obj, key, decode, sender_key)

    @staticmethod
    def parse_json(obj):
        """Parse JWE JSON Serialization.

        :param obj: JWE JSON Serialization as str or dict
        :return: Parsed JWE JSON Serialization as dict if `obj` is an str,
            or `obj` as is if `obj` is already a dict
        """
        return ensure_dict(obj, "JWE")

    def get_header_alg(self, header):
        if "alg" not in header:
            raise MissingAlgorithmError()

        alg = header["alg"]
        if self._algorithms is not None and alg not in self._algorithms:
            raise UnsupportedAlgorithmError()
        if alg not in self.ALG_REGISTRY:
            raise UnsupportedAlgorithmError()
        return self.ALG_REGISTRY[alg]

    def get_header_enc(self, header):
        if "enc" not in header:
            raise MissingEncryptionAlgorithmError()
        enc = header["enc"]
        if self._algorithms is not None and enc not in self._algorithms:
            raise UnsupportedEncryptionAlgorithmError()
        if enc not in self.ENC_REGISTRY:
            raise UnsupportedEncryptionAlgorithmError()
        return self.ENC_REGISTRY[enc]

    def get_header_zip(self, header):
        if "zip" in header:
            z = header["zip"]
            if self._algorithms is not None and z not in self._algorithms:
                raise UnsupportedCompressionAlgorithmError()
            if z not in self.ZIP_REGISTRY:
                raise UnsupportedCompressionAlgorithmError()
            return self.ZIP_REGISTRY[z]

    def _validate_sender_key(self, sender_key, alg):
        if isinstance(alg, JWEAlgorithmWithTagAwareKeyAgreement):
            if sender_key is None:
                raise ValueError(
                    f"{alg.name} algorithm requires sender_key but passed sender_key value is None"
                )
        else:
            if sender_key is not None:
                raise ValueError(
                    f"{alg.name} algorithm does not use sender_key but passed sender_key value is not None"
                )

    def _validate_private_headers(self, header, alg):
        # only validate private headers when developers set
        # private headers explicitly
        if self._private_headers is None:
            return

        names = self.REGISTERED_HEADER_PARAMETER_NAMES.copy()
        names = names.union(self._private_headers)

        if alg.EXTRA_HEADERS:
            names = names.union(alg.EXTRA_HEADERS)

        for k in header:
            if k not in names:
                raise InvalidHeaderParameterNameError(k)


def prepare_key(alg, header, key):
    if callable(key):
        key = key(header, None)
    elif key is None and "jwk" in header:
        key = header["jwk"]
    return alg.prepare_key(key)



import pytest
from httpx import ASGITransport
from starlette.config import Config
from starlette.requests import Request

from authlib.common.urls import url_decode
from authlib.common.urls import urlparse
from authlib.integrations.starlette_client import OAuth
from authlib.integrations.starlette_client import OAuthError

from ..asgi_helper import AsyncPathMapDispatch
from ..util import get_bearer_token


def test_register_remote_app():
    oauth = OAuth()
    with pytest.raises(AttributeError):
        assert oauth.dev.name == "dev"

    oauth.register(
        "dev",
        client_id="dev",
        client_secret="dev",
    )
    assert oauth.dev.name == "dev"
    assert oauth.dev.client_id == "dev"


def test_register_with_config():
    config = Config(environ={"DEV_CLIENT_ID": "dev"})
    oauth = OAuth(config)
    oauth.register("dev")
    assert oauth.dev.name == "dev"
    assert oauth.dev.client_id == "dev"


def test_register_with_overwrite():
    config = Config(environ={"DEV_CLIENT_ID": "dev"})
    oauth = OAuth(config)
    oauth.register("dev", client_id="not-dev", overwrite=True)
    assert oauth.dev.name == "dev"
    assert oauth.dev.client_id == "dev"


@pytest.mark.asyncio
async def test_oauth1_authorize():
    oauth = OAuth()
    transport = ASGITransport(
        AsyncPathMapDispatch(
            {
                "/request-token": {"body": "oauth_token=foo&oauth_verifier=baz"},
                "/token": {"body": "oauth_token=a&oauth_token_secret=b"},
            }
        )
    )
    client = oauth.register(
        "dev",
        client_id="dev",
        client_secret="dev",
        request_token_url="https://provider.test/request-token",
        api_base_url="https://resource.test/api",
        access_token_url="https://provider.test/token",
        authorize_url="https://provider.test/authorize",
        client_kwargs={
            "transport": transport,
        },
    )

    req_scope = {"type": "http", "session": {}}
    req = Request(req_scope)
    resp = await client.authorize_redirect(req, "https://client.test/callback")
    assert resp.status_code == 302
    url = resp.headers.get("Location")
    assert "oauth_token=foo" in url
    assert "_state_dev_foo" in req.session
    req.scope["query_string"] = "oauth_token=foo&oauth_verifier=baz"
    token = await client.authorize_access_token(req)
    assert token["oauth_token"] == "a"


@pytest.mark.asyncio
async def test_oauth2_authorize():
    oauth = OAuth()
    transport = ASGITransport(
        AsyncPathMapDispatch({"/token": {"body": get_bearer_token()}})
    )
    client = oauth.register(
        "dev",
        client_id="dev",
        client_secret="dev",
        api_base_url="https://resource.test/api",
        access_token_url="https://provider.test/token",
        authorize_url="https://provider.test/authorize",
        client_kwargs={
            "transport": transport,
        },
    )

    req_scope = {"type": "http", "session": {}}
    req = Request(req_scope)
    resp = await client.authorize_redirect(req, "https://client.test/callback")
    assert resp.status_code == 302
    url = resp.headers.get("Location")
    assert "state=" in url
    state = dict(url_decode(urlparse.urlparse(url).query))["state"]

    assert f"_state_dev_{state}" in req.session

    req_scope.update(
        {
            "path": "/",
            "query_string": f"code=a&state={state}",
            "session": req.session,
        }
    )
    req = Request(req_scope)
    token = await client.authorize_access_token(req)
    assert token["access_token"] == "a"


@pytest.mark.asyncio
async def test_oauth2_authorize_access_denied():
    oauth = OAuth()
    transport = ASGITransport(
        AsyncPathMapDispatch({"/token": {"body": get_bearer_token()}})
    )
    client = oauth.register(
        "dev",
        client_id="dev",
        client_secret="dev",
        api_base_url="https://resource.test/api",
        access_token_url="https://provider.test/token",
        authorize_url="https://provider.test/authorize",
        client_kwargs={
            "transport": transport,
        },
    )

    req = Request(
        {
            "type": "http",
            "session": {},
            "path": "/",
            "query_string": "error=access_denied&error_description=Not+Allowed",
        }
    )
    with pytest.raises(OAuthError):
        await client.authorize_access_token(req)


@pytest.mark.asyncio
async def test_oauth2_authorize_code_challenge():
    transport = ASGITransport(
        AsyncPathMapDispatch({"/token": {"body": get_bearer_token()}})
    )
    oauth = OAuth()
    client = oauth.register(
        "dev",
        client_id="dev",
        api_base_url="https://resource.test/api",
        access_token_url="https://provider.test/token",
        authorize_url="https://provider.test/authorize",
        client_kwargs={
            "code_challenge_method": "S256",
            "transport": transport,
        },
    )

    req_scope = {"type": "http", "session": {}}
    req = Request(req_scope)

    resp = await client.authorize_redirect(
        req, redirect_uri="https://client.test/callback"
    )
    assert resp.status_code == 302

    url = resp.headers.get("Location")
    assert "code_challenge=" in url
    assert "code_challenge_method=S256" in url

    state = dict(url_decode(urlparse.urlparse(url).query))["state"]
    state_data = req.session[f"_state_dev_{state}"]["data"]

    verifier = state_data["code_verifier"]
    assert verifier is not None

    req_scope.update(
        {
            "path": "/",
            "query_string": f"code=a&state={state}".encode(),
            "session": req.session,
        }
    )
    req = Request(req_scope)

    token = await client.authorize_access_token(req)
    assert token["access_token"] == "a"


@pytest.mark.asyncio
async def test_with_fetch_token_in_register():
    async def fetch_token(request):
        return {"access_token": "dev", "token_type": "bearer"}

    transport = ASGITransport(AsyncPathMapDispatch({"/user": {"body": {"sub": "123"}}}))
    oauth = OAuth()
    client = oauth.register(
        "dev",
        client_id="dev",
        client_secret="dev",
        api_base_url="https://resource.test/api",
        access_token_url="https://provider.test/token",
        authorize_url="https://provider.test/authorize",
        fetch_token=fetch_token,
        client_kwargs={
            "transport": transport,
        },
    )

    req_scope = {"type": "http", "session": {}}
    req = Request(req_scope)
    resp = await client.get("/user", request=req)
    assert resp.json()["sub"] == "123"


@pytest.mark.asyncio
async def test_with_fetch_token_in_oauth():
    async def fetch_token(name, request):
        return {"access_token": "dev", "token_type": "bearer"}

    transport = ASGITransport(AsyncPathMapDispatch({"/user": {"body": {"sub": "123"}}}))
    oauth = OAuth(fetch_token=fetch_token)
    client = oauth.register(
        "dev",
        client_id="dev",
        client_secret="dev",
        api_base_url="https://resource.test/api",
        access_token_url="https://provider.test/token",
        authorize_url="https://provider.test/authorize",
        client_kwargs={
            "transport": transport,
        },
    )

    req_scope = {"type": "http", "session": {}}
    req = Request(req_scope)
    resp = await client.get("/user", request=req)
    assert resp.json()["sub"] == "123"


@pytest.mark.asyncio
async def test_request_withhold_token():
    oauth = OAuth()
    transport = ASGITransport(AsyncPathMapDispatch({"/user": {"body": {"sub": "123"}}}))
    client = oauth.register(
        "dev",
        client_id="dev",
        client_secret="dev",
        api_base_url="https://resource.test/api",
        access_token_url="https://provider.test/token",
        authorize_url="https://provider.test/authorize",
        client_kwargs={
            "transport": transport,
        },
    )
    req_scope = {"type": "http", "session": {}}
    req = Request(req_scope)
    resp = await client.get("/user", request=req, withhold_token=True)
    assert resp.json()["sub"] == "123"


@pytest.mark.asyncio
async def test_oauth2_authorize_no_url():
    oauth = OAuth()
    client = oauth.register(
        "dev",
        client_id="dev",
        client_secret="dev",
        api_base_url="https://resource.test/api",
        access_token_url="https://provider.test/token",
    )
    req_scope = {"type": "http", "session": {}}
    req = Request(req_scope)
    with pytest.raises(RuntimeError):
        await client.create_authorization_url(req)


@pytest.mark.asyncio
async def test_oauth2_authorize_with_metadata():
    oauth = OAuth()
    transport = ASGITransport(
        AsyncPathMapDispatch(
            {
                "/.well-known/openid-configuration": {
                    "body": {
                        "authorization_endpoint": "https://provider.test/authorize"
                    }
                }
            }
        )
    )
    client = oauth.register(
        "dev",
        client_id="dev",
        client_secret="dev",
        api_base_url="https://resource.test/api",
        access_token_url="https://provider.test/token",
        server_metadata_url="https://provider.test/.well-known/openid-configuration",
        client_kwargs={
            "transport": transport,
        },
    )
    req_scope = {"type": "http", "session": {}}
    req = Request(req_scope)
    resp = await client.authorize_redirect(req, "https://client.test/callback")
    assert resp.status_code == 302


@pytest.mark.asyncio
async def test_oauth2_authorize_form_post_callback():
    """Test that POST callbacks (form_post response mode) work properly."""
    oauth = OAuth()
    transport = ASGITransport(
        AsyncPathMapDispatch({"/token": {"body": get_bearer_token()}})
    )
    client = oauth.register(
        "dev",
        client_id="dev",
        client_secret="dev",
        api_base_url="https://resource.test/api",
        access_token_url="https://provider.test/token",
        authorize_url="https://provider.test/authorize",
        client_kwargs={
            "transport": transport,
        },
    )

    req = Request({"type": "http", "session": {}})
    resp = await client.authorize_redirect(req, "https://client.test/callback")
    url = resp.headers.get("Location")
    state = dict(url_decode(urlparse.urlparse(url).query))["state"]

    req_scope_post = {
        "type": "http",
        "method": "POST",
        "path": "/callback",
        "query_string": b"",
        "headers": [(b"content-type", b"application/x-www-form-urlencoded")],
        "session": req.session,
    }

    async def receive():
        return {
            "type": "http.request",
            "body": f"code=test_code&state={state}".encode(),
        }

    req_post = Request(req_scope_post, receive=receive)

    token = await client.authorize_access_token(req_post)
    assert token["access_token"] == "a"



import pytest
from flask import json

from authlib.jose import jwt
from authlib.oauth2.rfc7591 import (
    ClientRegistrationEndpoint as _ClientRegistrationEndpoint,
)
from tests.util import read_file_path

from .models import Client
from .models import db


class ClientRegistrationEndpoint(_ClientRegistrationEndpoint):
    software_statement_alg_values_supported = ["RS256"]

    def authenticate_token(self, request):
        auth_header = request.headers.get("Authorization")
        if auth_header:
            request.user_id = 1
            return auth_header

    def resolve_public_key(self, request):
        return read_file_path("rsa_public.pem")

    def save_client(self, client_info, client_metadata, request):
        client = Client(user_id=request.user_id, **client_info)
        client.set_client_metadata(client_metadata)
        db.session.add(client)
        db.session.commit()
        return client


@pytest.fixture
def metadata():
    return {}


@pytest.fixture(autouse=True)
def server(server, app, metadata):
    class MyClientRegistration(ClientRegistrationEndpoint):
        def get_server_metadata(test_client):
            return metadata

    server.register_endpoint(MyClientRegistration)

    @app.route("/create_client", methods=["POST"])
    def create_client():
        return server.create_endpoint_response("client_registration")

    return server


def test_access_denied(test_client):
    rv = test_client.post("/create_client", json={})
    resp = json.loads(rv.data)
    assert resp["error"] == "access_denied"


def test_invalid_request(test_client):
    headers = {"Authorization": "bearer abc"}
    rv = test_client.post("/create_client", json={}, headers=headers)
    resp = json.loads(rv.data)
    assert resp["error"] == "invalid_request"


def test_create_client(test_client):
    headers = {"Authorization": "bearer abc"}
    body = {"client_name": "Authlib"}
    rv = test_client.post("/create_client", json=body, headers=headers)
    resp = json.loads(rv.data)
    assert "client_id" in resp
    assert resp["client_name"] == "Authlib"


def test_software_statement(test_client):
    payload = {"software_id": "uuid-123", "client_name": "Authlib"}
    s = jwt.encode({"alg": "RS256"}, payload, read_file_path("rsa_private.pem"))
    body = {
        "software_statement": s.decode("utf-8"),
    }

    headers = {"Authorization": "bearer abc"}
    rv = test_client.post("/create_client", json=body, headers=headers)
    resp = json.loads(rv.data)
    assert "client_id" in resp
    assert resp["client_name"] == "Authlib"


def test_no_public_key(test_client, server):
    class ClientRegistrationEndpoint2(ClientRegistrationEndpoint):
        def get_server_metadata(test_client):
            return None

        def resolve_public_key(self, request):
            return None

    payload = {"software_id": "uuid-123", "client_name": "Authlib"}
    s = jwt.encode({"alg": "RS256"}, payload, read_file_path("rsa_private.pem"))
    body = {
        "software_statement": s.decode("utf-8"),
    }

    server._endpoints[ClientRegistrationEndpoint.ENDPOINT_NAME] = [
        ClientRegistrationEndpoint2(server)
    ]

    headers = {"Authorization": "bearer abc"}
    rv = test_client.post("/create_client", json=body, headers=headers)
    resp = json.loads(rv.data)
    assert resp["error"] in "unapproved_software_statement"


def test_scopes_supported(test_client, metadata):
    metadata["scopes_supported"] = ["profile", "email"]

    headers = {"Authorization": "bearer abc"}
    body = {"scope": "profile email", "client_name": "Authlib"}
    rv = test_client.post("/create_client", json=body, headers=headers)
    resp = json.loads(rv.data)
    assert "client_id" in resp
    assert resp["client_name"] == "Authlib"

    body = {"scope": "profile email address", "client_name": "Authlib"}
    rv = test_client.post("/create_client", json=body, headers=headers)
    resp = json.loads(rv.data)
    assert resp["error"] in "invalid_client_metadata"


def test_response_types_supported(test_client, metadata):
    metadata["response_types_supported"] = ["code", "code id_token"]

    headers = {"Authorization": "bearer abc"}
    body = {"response_types": ["code"], "client_name": "Authlib"}
    rv = test_client.post("/create_client", json=body, headers=headers)
    resp = json.loads(rv.data)
    assert "client_id" in resp
    assert resp["client_name"] == "Authlib"

    # The items order should not matter
    # Extension response types MAY contain a space-delimited (%x20) list of
    # values, where the order of values does not matter (e.g., response
    # type "a b" is the same as "b a").
    headers = {"Authorization": "bearer abc"}
    body = {"response_types": ["id_token code"], "client_name": "Authlib"}
    rv = test_client.post("/create_client", json=body, headers=headers)
    resp = json.loads(rv.data)
    assert "client_id" in resp
    assert resp["client_name"] == "Authlib"

    # https://www.rfc-editor.org/rfc/rfc7591.html#section-2
    # If omitted, the default is that the client will use only the "code"
    # response type.
    body = {"client_name": "Authlib"}
    rv = test_client.post("/create_client", json=body, headers=headers)
    resp = json.loads(rv.data)
    assert "client_id" in resp
    assert resp["client_name"] == "Authlib"

    body = {"response_types": ["code", "token"], "client_name": "Authlib"}
    rv = test_client.post("/create_client", json=body, headers=headers)
    resp = json.loads(rv.data)
    assert resp["error"] in "invalid_client_metadata"


def test_grant_types_supported(test_client, metadata):
    metadata["grant_types_supported"] = ["authorization_code", "password"]

    headers = {"Authorization": "bearer abc"}
    body = {"grant_types": ["password"], "client_name": "Authlib"}
    rv = test_client.post("/create_client", json=body, headers=headers)
    resp = json.loads(rv.data)
    assert "client_id" in resp
    assert resp["client_name"] == "Authlib"

    # https://www.rfc-editor.org/rfc/rfc7591.html#section-2
    # If omitted, the default behavior is that the client will use only
    # the "authorization_code" Grant Type.
    body = {"client_name": "Authlib"}
    rv = test_client.post("/create_client", json=body, headers=headers)
    resp = json.loads(rv.data)
    assert "client_id" in resp
    assert resp["client_name"] == "Authlib"

    body = {"grant_types": ["client_credentials"], "client_name": "Authlib"}
    rv = test_client.post("/create_client", json=body, headers=headers)
    resp = json.loads(rv.data)
    assert resp["error"] in "invalid_client_metadata"


def test_token_endpoint_auth_methods_supported(test_client, metadata):
    metadata["token_endpoint_auth_methods_supported"] = ["client_secret_basic"]

    headers = {"Authorization": "bearer abc"}
    body = {
        "token_endpoint_auth_method": "client_secret_basic",
        "client_name": "Authlib",
    }
    rv = test_client.post("/create_client", json=body, headers=headers)
    resp = json.loads(rv.data)
    assert "client_id" in resp
    assert resp["client_name"] == "Authlib"

    body = {"token_endpoint_auth_method": "none", "client_name": "Authlib"}
    rv = test_client.post("/create_client", json=body, headers=headers)
    resp = json.loads(rv.data)
    assert resp["error"] in "invalid_client_metadata"



import os

import pytest
from flask import Flask


@pytest.fixture(autouse=True)
def env():
    os.environ["AUTHLIB_INSECURE_TRANSPORT"] = "true"
    yield
    del os.environ["AUTHLIB_INSECURE_TRANSPORT"]


@pytest.fixture
def app():
    app = Flask(__name__)
    app.debug = True
    app.testing = True
    app.secret_key = "testing"
    app.config.update(
        {
            "OAUTH1_SUPPORTED_SIGNATURE_METHODS": [
                "PLAINTEXT",
                "HMAC-SHA1",
                "RSA-SHA1",
            ],
            "SQLALCHEMY_TRACK_MODIFICATIONS": False,
            "SQLALCHEMY_DATABASE_URI": "sqlite://",
        }
    )

    with app.app_context():
        yield app


@pytest.fixture
def test_client(app, db):
    return app.test_client()


@pytest.fixture
def db(app):
    from .oauth1_server import db

    db.init_app(app)
    db.create_all()
    yield db
    db.drop_all()



from django.conf import settings
from django.test import RequestFactory
from django.utils.module_loading import import_module


class RequestClient(RequestFactory):
    @property
    def session(self):
        engine = import_module(settings.SESSION_ENGINE)
        cookie = self.cookies.get(settings.SESSION_COOKIE_NAME)
        if cookie:
            return engine.SessionStore(cookie.value)

        session = engine.SessionStore()
        session.save()
        self.cookies[settings.SESSION_COOKIE_NAME] = session.session_key
        return session



import os

import pytest

from authlib.integrations.django_oauth2 import AuthorizationServer

from .models import Client
from .models import OAuth2Token
from .models import User

pytestmark = pytest.mark.django_db


@pytest.fixture(autouse=True)
def env():
    os.environ["AUTHLIB_INSECURE_TRANSPORT"] = "true"
    yield
    os.environ.pop("AUTHLIB_INSECURE_TRANSPORT", None)


@pytest.fixture(autouse=True)
def server(settings):
    settings.AUTHLIB_OAUTH2_PROVIDER = {}
    return AuthorizationServer(Client, OAuth2Token)


@pytest.fixture(autouse=True)
def user(db):
    user = User(username="foo")
    user.set_password("ok")
    user.save()
    yield user
    user.delete()


@pytest.fixture
def token(user):
    token = OAuth2Token(
        user_id=user.pk,
        client_id="client-id",
        token_type="bearer",
        access_token="a1",
        refresh_token="r1",
        scope="profile",
        expires_in=3600,
    )
    token.save()
    yield token
    token.delete()
